#!/usr/bin/env python

import os
import platform
from setuptools import Extension, setup, find_packages
from glob import glob

if os.name == "nt":
    raise NotImplementedError(
        "Python metacells does not support native windows.\nInstead, "
        "install Windows Subsystem for Linux, and metacells within it."
    )

DEFINE_MACROS = [("ASSERT_LEVEL", 1)]  # 0 for none, 1 for fast, 2 for slow.
COMPILE_ARGS = [f"-I{os.getcwd()}", "-std=c++14"]
# COMPILE_ARGS += ["-fsanitize=address", "-fno-omit-frame-pointer"]
# COMPILE_ARGS += ["-fopt-info-vec-all", "-fopt-info-loop-optimized"]

with open("metacells/should_check_avx2.py", "w") as file:
    file.write("# file generated by setup\n")
    file.write("# don't change, don't track in version control\n")
    if str(os.getenv("WHEEL", "")) == "":
        COMPILE_ARGS += ["-march=native", "-mtune=native"]
        file.write("SHOULD_CHECK_AVX2 = False\n")
    elif platform.processor() == "x86_64":
        COMPILE_ARGS += ["-march=haswell", "-mtune=broadwell"]
        DEFINE_MACROS.append(("USE_AVX2", 1))
        file.write("SHOULD_CHECK_AVX2 = True\n")
    else:
        file.write("SHOULD_CHECK_AVX2 = False\n")

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = open("requirements.txt").read().split()
test_requirements = open("requirements_test.txt").read().split()
dev_requirements = open("requirements_dev.txt").read().split()

packages = ['metacells'] + [sub_dir for sub_dir in glob('metacells/*') if '.' not in sub_dir and '__' not in sub_dir]

setup(
    author="Oren Ben-Kiki",
    author_email="oren@ben-kiki.org",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    description="Single-cell RNA Sequencing Analysis",
    headers=["metacells/extensions.h"],
    ext_modules=[
        Extension(  #
            "metacells.extensions",
            include_dirs=["pybind11/include"],
            sources=[
                "metacells/auroc.cpp",
                "metacells/choose_seeds.cpp",
                "metacells/correlate.cpp",
                "metacells/cover.cpp",
                "metacells/downsample.cpp",
                "metacells/extensions.cpp",
                "metacells/folds.cpp",
                "metacells/gaps.cpp",
                "metacells/logistics.cpp",
                "metacells/partitions.cpp",
                "metacells/prune_per.cpp",
                "metacells/rank.cpp",
                "metacells/relayout.cpp",
                "metacells/shuffle.cpp",
                "metacells/top_per.cpp",
            ],
            define_macros=DEFINE_MACROS,
            extra_compile_args=COMPILE_ARGS,
        ),
    ],
    entry_points={
        "console_scripts": [
            "metacells_timing=metacells.scripts.timing:main",
        ]
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="metacells",
    name="metacells",
    packages=packages,
    test_suite="tests",
    tests_require=test_requirements,
    extras_require={"dev": dev_requirements},
    url="https://github.com/tanaylab/metacells.git",
    version="0.9.3",
)



"""
Related Genes
-------------
"""

from typing import List
from typing import Tuple
from typing import Union

import fastcluster as fc  # type: ignore
import numpy as np
import pandas as pd  # type: ignore
import scipy.sparse as sp  # type: ignore
import scipy.spatial.distance as scd  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

__all__ = [
    "relate_to_lateral_genes",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def relate_to_lateral_genes(  # pylint: disable=too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    max_sampled_cells: int = pr.related_max_sampled_cells,
    genes_similarity_method: str = pr.related_genes_similarity_method,
    genes_cluster_method: str = pr.related_genes_cluster_method,
    max_genes_of_modules: int = pr.related_max_genes_of_modules,
    min_mean_gene_fraction: float = pr.related_min_mean_gene_fraction,
    min_gene_correlation: float = pr.related_min_gene_correlation,
    random_seed: int = 0,
) -> None:
    """
    Detect coarse relations between genes and lateral genes based on ``what`` (default: {what}) data.

    This is a quick-and-dirty way to look for genes highly correlated with lateral genes.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where ``what`` is a
    per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix. The data should contain a ``lateral_gene`` mask containing some known-to-be lateral genes.

    **Returns**

    Variable-pair (Gene) Annotations
        ``lateral_genes_similarity``
            The similarity between each genes related to lateral genes.

    Variable (Gene) Annotations
        ``lateral_genes_module``
            The index of the related to lateral gene module for each gene.

    **Computation Parameters**

    1. If we have more than ``max_sampled_cells`` (default: {max_sampled_cells}), pick this number
       of random cells from the data using the ``random_seed``.

    2. Start by marking as "related genes" all the genes listed in ``lateral_gene``.

    3. Pick as candidates genes whose mean fraction in the population is at least the (very low)
       ``min_mean_gene_fraction`` (default: {min_mean_gene_fraction}) (and that are not lateral genes).

    4. Compute the correlation between the related genes and the candidate genes. Move from the candidate genes to the
       lateral genes any genes whose correlation is at least ``min_gene_correlation`` (default: {min_gene_correlation}).

    5. Repeat step 4 until no additional related genes are found.

    6. Compute the similarity between the all related genes using
       :py:func:`metacells.tools.similarity.compute_var_var_similarity` using the ``genes_similarity_method`` (default:
       {genes_similarity_method}).

    7. Create a hierarchical clustering of the candidate genes using the ``genes_cluster_method``
       (default: {genes_cluster_method}).

    8. Identify gene modules in the hierarchical clustering by bottom-up combining genes as long as we have at most
       ``max_genes_of_modules`` (default: {max_genes_of_modules}). Note that this may leave modules with a smaller
       number of genes, and modules that do not contain any lateral genes.
    """
    if max_sampled_cells < adata.n_obs:
        np.random.seed(random_seed)
        cell_indices = np.random.choice(np.arange(adata.n_obs), size=max_sampled_cells, replace=False)
        sdata = ut.slice(adata, obs=cell_indices, name=".sampled", top_level=False)
    else:
        sdata = ut.copy_adata(adata, top_level=False)

    umis_per_cell_per_gene = ut.get_vo_proper(sdata, what)
    fraction_per_gene_per_cell = ut.fraction_by(umis_per_cell_per_gene, by="row")
    fraction_per_gene_per_cell = ut.to_layout(fraction_per_gene_per_cell, layout="column_major")

    mean_fraction_per_gene = ut.mean_per(fraction_per_gene_per_cell, per="column")
    candidate_genes_mask = mean_fraction_per_gene >= min_mean_gene_fraction
    ut.log_calc("candidate_genes_mask", candidate_genes_mask)

    related_genes_mask = np.zeros(sdata.n_vars, dtype="bool")
    additional_genes_mask = ut.get_v_numpy(sdata, "lateral_gene").copy()

    while np.any(additional_genes_mask):
        related_genes_mask |= additional_genes_mask
        ut.log_calc("related_genes_mask", related_genes_mask)

        candidate_genes_mask &= ~additional_genes_mask
        ut.log_calc("candidate_genes_mask", candidate_genes_mask)
        if not np.any(candidate_genes_mask):
            break

        fraction_per_candidate_gene_per_cell = ut.to_numpy_matrix(fraction_per_gene_per_cell[:, candidate_genes_mask])
        fraction_per_additional_gene_per_cell = ut.to_numpy_matrix(fraction_per_gene_per_cell[:, additional_genes_mask])

        fraction_per_cell_per_candidate_gene = np.transpose(fraction_per_candidate_gene_per_cell)
        fraction_per_cell_per_additional_gene = np.transpose(fraction_per_additional_gene_per_cell)

        fraction_per_cell_per_candidate_gene = ut.to_layout(fraction_per_cell_per_candidate_gene, layout="row_major")
        fraction_per_cell_per_additional_gene = ut.to_layout(fraction_per_cell_per_additional_gene, layout="row_major")

        correlation_per_related_per_candidate = ut.cross_corrcoef_rows(
            fraction_per_cell_per_candidate_gene,
            fraction_per_cell_per_additional_gene,
            reproducible=random_seed != 0,
        )
        max_correlation_per_candidate = ut.max_per(correlation_per_related_per_candidate, per="row")
        assert len(max_correlation_per_candidate) == np.sum(candidate_genes_mask)

        additional_genes_mask = np.zeros(sdata.n_vars, dtype="bool")
        additional_genes_mask[candidate_genes_mask] = max_correlation_per_candidate >= min_gene_correlation
        ut.log_calc("additional_genes_mask", additional_genes_mask)

    name = (ut.get_name(sdata) or "cells") + ".related"
    var_names = sdata.var_names[related_genes_mask]
    sdata = AnnData(fraction_per_gene_per_cell[:, related_genes_mask])
    sdata.var_names = var_names
    ut.set_name(sdata, name)

    frame = tl.compute_var_var_similarity(
        sdata, what, method=genes_similarity_method, reproducible=(random_seed != 0), inplace=False
    )
    assert frame is not None
    similarity = ut.to_layout(ut.to_numpy_matrix(frame), layout="row_major")
    ut.log_calc("similarity", similarity)

    linkage = _cluster_genes(similarity, genes_cluster_method)
    ut.log_calc("linkage")
    clusters = _linkage_to_clusters(linkage, max_genes_of_modules, sdata.n_vars)
    ut.log_calc("cluster")

    cluster_of_genes = pd.Series(np.full(adata.n_vars, -1, dtype="int32"), index=adata.var_names)
    for cluster_index, gene_indices in enumerate(clusters):
        cluster_of_genes[sdata.var_names[gene_indices]] = cluster_index

    ut.set_v_data(adata, "lateral_genes_module", cluster_of_genes, formatter=ut.groups_description)

    related_gene_indices = np.where(related_genes_mask)[0]
    data = similarity.flatten(order="C")
    rows = np.repeat(related_gene_indices, len(related_gene_indices))
    columns = np.tile(related_gene_indices, len(related_gene_indices))
    full_similarity = sp.csr_matrix((data, (rows, columns)), shape=(adata.n_vars, adata.n_vars))

    ut.set_vv_data(adata, "lateral_genes_similarity", full_similarity)


# TODO: Replicated in metacell.tools.rare
@ut.timed_call()
def _cluster_genes(
    similarities_between_candidate_genes: ut.NumpyMatrix,
    genes_cluster_method: str,
) -> List[Tuple[int, int]]:
    with ut.timed_step("scipy.pdist"):
        ut.timed_parameters(size=similarities_between_candidate_genes.shape[0])
        distances = scd.pdist(similarities_between_candidate_genes)

    with ut.timed_step("scipy.linkage"):
        ut.timed_parameters(size=distances.shape[0], method=genes_cluster_method)
        linkage = fc.linkage(distances, method=genes_cluster_method)

    return linkage


@ut.timed_call()
def _linkage_to_clusters(
    linkage: List[Tuple[int, int]],
    max_entries_of_modules: int,
    entries_count: int,
) -> List[List[int]]:
    entries_of_cluster = {index: [index] for index in range(entries_count)}

    for link_index, link_data in enumerate(linkage):
        link_index += entries_count

        left_index = int(link_data[0])
        right_index = int(link_data[1])

        left_entries = entries_of_cluster.get(left_index)
        right_entries = entries_of_cluster.get(right_index)

        if (
            left_entries is None
            or right_entries is None
            or len(left_entries) + len(right_entries) > max_entries_of_modules
        ):
            continue

        entries_of_cluster[link_index] = sorted(left_entries + right_entries)
        del entries_of_cluster[left_index]
        del entries_of_cluster[right_index]

    return list(entries_of_cluster.values())



"""
UMAP
----

It is useful to project the metacells data to a 2D scatter plot (where each point is a metacell).
The steps provided here are expected to yield a reasonable such projection, though as always you
might need to tweak the parameters or even the overall flow for specific data sets.
"""

from re import Pattern
from typing import Collection
from typing import Optional
from typing import Tuple
from typing import Union

import igraph as ig  # type: ignore
import numpy as np
from anndata import AnnData  # type: ignore
from scipy import sparse  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

__all__ = [
    "compute_knn_by_markers",
    "compute_umap_by_markers",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_knn_by_markers(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    marker_gene_names: Optional[Collection[str]] = None,
    marker_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    max_marker_genes: Optional[int] = pr.umap_max_marker_genes,
    ignore_lateral_genes: bool = pr.umap_ignore_lateral_genes,
    similarity_value_regularization: float = pr.umap_similarity_value_regularization,
    similarity_log_data: bool = pr.umap_similarity_log_data,
    similarity_method: str = pr.umap_similarity_method,
    logistics_location: float = pr.logistics_location,
    logistics_slope: float = pr.logistics_slope,
    k: int,
    balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    min_outgoing_degree: int = pr.markers_knn_min_outgoing_degree,
    reproducible: bool,
) -> ut.PandasFrame:
    """
    Compute KNN graph between metacells based on marker genes.

    If ``reproducible`` is ``True``, a slower (still parallel) but
    reproducible algorithm will be used to compute Pearson correlations.

    **Input**

    Annotated ``adata`` where each observation is a metacells and the variables are genes, are genes, where ``what`` is
    a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix. Should contain a ``marker_gene`` mask unless explicitly specifying the marker genes.

    **Returns**

    Sets the following in ``adata``:

    Observations-Pair (Metacells) Annotations
        ``obs_outgoing_weights``
            A sparse square matrix where each non-zero entry is the weight of an edge between a pair
            of cells or genes, where the sum of the weights of the outgoing edges for each element
            is 1 (there is always at least one such edge).

    Also return a pandas data frame of the similarities between the observations (metacells).

    **Computation Parameters**

    1. If ``marker_gene_names`` and/or ``marker_gene_patterns`` were specified, use the matching genes.
       Otherwise, use the ``marker_gene`` mask.

    2. If ``ignore_lateral_genes`` (default: {ignore_lateral_genes}), then remove any genes marked as lateral
       from the mask.

    3. If ``max_marker_genes`` (default: {max_marker_genes}) is not ``None``, then pick this number
       of marker genes with the highest variance.

    4. Compute the fractions of each ``marker_gene`` in each cell, and add the
       ``similarity_value_regularization`` (default: {similarity_value_regularization}) to it.

    5. If ``similarity_log_data`` (default: {similarity_log_data}), invoke the
       :py:func:`metacells.utilities.computation.log_data` function to compute the log (base 2) of
       the data.

    6. Invoke :py:func:`metacells.tools.similarity.compute_obs_obs_similarity` using
       ``similarity_method`` (default: {similarity_method}), ``logistics_location`` (default:
       {logistics_slope}) and ``logistics_slope`` (default: {logistics_slope}) and convert this
       to distances.

    7. Invoke :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph` using the distances,
       ``k`` (no default!), ``balanced_ranks_factor`` (default: {balanced_ranks_factor}),
       ``incoming_degree_factor`` (default: {incoming_degree_factor}), ``outgoing_degree_factor``
       (default: {outgoing_degree_factor}) and ``min_outgoing_degree`` (default: {min_outgoing_degree})
       to compute a "skeleton" graph to overlay on top of the UMAP graph.
    """
    assert max_marker_genes is None or max_marker_genes > 0
    if marker_gene_names is None and marker_gene_patterns is None:
        marker_genes_mask = ut.get_v_numpy(adata, "marker_gene")
    else:
        marker_genes_series = tl.find_named_genes(adata, names=marker_gene_names, patterns=marker_gene_patterns)
        assert marker_genes_series is not None
        marker_genes_mask = ut.to_numpy_vector(marker_genes_series)

    if ignore_lateral_genes:
        lateral_genes_mask = ut.get_v_numpy(adata, "lateral_gene")
        marker_genes_mask = marker_genes_mask & ~lateral_genes_mask
        ut.log_calc("marker_genes_mask", marker_genes_mask)

    all_fractions = ut.get_vo_proper(adata, what, layout="row_major")

    index_per_marker_gene = np.where(marker_genes_mask)[0]
    fraction_per_metacell_per_marker_gene = ut.to_numpy_matrix(all_fractions[:, index_per_marker_gene])

    if max_marker_genes is not None and max_marker_genes < len(index_per_marker_gene):
        fraction_per_metacell_per_marker_gene = ut.to_layout(
            fraction_per_metacell_per_marker_gene, layout="column_major"
        )
        variance_per_marker_gene = ut.variance_per(fraction_per_metacell_per_marker_gene, per="column")
        variance_per_marker_gene *= -1
        chosen_positions = np.argpartition(variance_per_marker_gene, max_marker_genes)[:max_marker_genes]
        ut.log_calc("chosen_positions", chosen_positions)
        index_per_marker_gene = index_per_marker_gene[chosen_positions]
        fraction_per_metacell_per_marker_gene = fraction_per_metacell_per_marker_gene[:, chosen_positions]

    fraction_per_metacell_per_marker_gene = ut.to_layout(fraction_per_metacell_per_marker_gene, layout="row_major")
    fraction_per_metacell_per_marker_gene += similarity_value_regularization

    if similarity_log_data:
        fraction_per_metacell_per_marker_gene = ut.log_data(fraction_per_metacell_per_marker_gene, base=2)

    tdata = ut.slice(adata, vars=index_per_marker_gene)
    similarities = tl.compute_obs_obs_similarity(
        tdata,
        fraction_per_metacell_per_marker_gene,
        method=similarity_method,
        reproducible=reproducible,
        logistics_location=logistics_location,
        logistics_slope=logistics_slope,
        inplace=False,
    )
    assert similarities is not None

    tl.compute_obs_obs_knn_graph(
        adata,
        similarities,
        k=k,
        balanced_ranks_factor=balanced_ranks_factor,
        incoming_degree_factor=incoming_degree_factor,
        outgoing_degree_factor=outgoing_degree_factor,
        min_outgoing_degree=min_outgoing_degree,
    )

    return similarities


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_umap_by_markers(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    marker_gene_names: Optional[Collection[str]] = None,
    marker_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    max_marker_genes: Optional[int] = pr.umap_max_marker_genes,
    ignore_lateral_genes: bool = pr.umap_ignore_lateral_genes,
    similarity_value_regularization: float = pr.umap_similarity_value_regularization,
    similarity_log_data: bool = pr.umap_similarity_log_data,
    similarity_method: str = pr.umap_similarity_method,
    logistics_location: float = pr.logistics_location,
    logistics_slope: float = pr.logistics_slope,
    skeleton_k: int = pr.skeleton_k,
    balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    min_outgoing_degree: int = pr.markers_knn_min_outgoing_degree,
    umap_k: int = pr.umap_k,
    dimensions: int = 2,
    min_dist: float = pr.umap_min_dist,
    spread: float = pr.umap_spread,
    random_seed: int,
) -> None:
    """
    Compute a UMAP projection of the (meta)cells.

    **Input**

    Annotated ``adata`` where each observation is a metacells and the variables are genes, are genes, where ``what`` is
    a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix. Should contain a ``marker_gene`` mask unless explicitly specifying the marker genes.

    **Returns**

    Sets the following annotations in ``adata``:

    Observation-Observation (Metacell-Metacell) Annotations
        ``umap_distances``
            A sparse symmetric matrix of the graph of distances between the metacells.

    Observation (Metacell) Annotations
        ``x``, ``y`` (if ``dimensions`` is 2)
            The X and Y coordinates of each metacell in the UMAP projection.

        ``u``, ``v``, ``w`` (if ``dimensions`` is 3)
            The U, V, W coordinates of each metacell in the UMAP projection.

    **Computation Parameters**

    1. Invoke :py:func:`metacells.pipeline.umap.compute_knn_by_markers` using
       ``marker_gene_names`` (default: {marker_gene_names}), ``marker_gene_patterns`` (default: {marker_gene_patterns}),
       ``max_marker_genes`` (default: {max_marker_genes}), ``ignore_lateral_genes`` (default: {ignore_lateral_genes}),
       ``similarity_value_regularization`` (default: {similarity_value_regularization}), ``similarity_log_data``
       (default: {similarity_log_data}), ``similarity_method`` (default: {similarity_method}), ``logistics_location``
       (default: {logistics_location}), ``logistics_slope`` (default: {logistics_slope}), ``skeleton_k`` (default:
       {skeleton_k}), ``balanced_ranks_factor`` (default: {balanced_ranks_factor}), ``incoming_degree_factor`` (default:
       {incoming_degree_factor}) and ``outgoing_degree_factor`` (default: {outgoing_degree_factor}) and
       ``min_outgoing_degree`` (default: {min_outgoing_degree}) to compute a "skeleton" graph to overlay on top of the
       UMAP graph. Specify a non-zero ``random_seed`` to make this reproducible.

    2. Invoke :py:func:`metacells.tools.layout.umap_by_distances` using the distances, ``umap_k``
       (default: {umap_k}), ``min_dist`` (default: {min_dist}), ``spread`` (default: {spread}),
       dimensions (default: {dimensions}) and ``random_seed``.

    .. note::

        Keep in mind that the KNN graph used by UMAP (controlled by ``umap_k``) is *not* identical to the KNN graph we
        compute (controlled by ``skeleton_k``). By default, we choose ``skeleton_k < umap_k``, as the purpose of the
        skeleton KNN is to highlight the "strong" structure of the data; in practice this strong skeleton is highly
        compatible with the structure used by UMAP, so it serves it purpose reasonably well. It would have been nice to
        make these compatible, but UMAP is not friendly towards dictating a KNN graph from the outside.
    """
    similarities = compute_knn_by_markers(
        adata,
        what,
        marker_gene_names=marker_gene_names,
        marker_gene_patterns=marker_gene_patterns,
        max_marker_genes=max_marker_genes,
        ignore_lateral_genes=ignore_lateral_genes,
        similarity_value_regularization=similarity_value_regularization,
        similarity_log_data=similarity_log_data,
        similarity_method=similarity_method,
        logistics_location=logistics_location,
        logistics_slope=logistics_slope,
        k=skeleton_k,
        balanced_ranks_factor=balanced_ranks_factor,
        incoming_degree_factor=incoming_degree_factor,
        outgoing_degree_factor=outgoing_degree_factor,
        min_outgoing_degree=min_outgoing_degree,
        reproducible=(random_seed != 0),
    )

    distances = ut.to_numpy_matrix(similarities)
    distances *= -1
    distances += 1
    np.fill_diagonal(distances, 0.0)
    distances = sparse.csr_matrix(distances)

    ut.set_oo_data(adata, "umap_distances", distances)

    tl.umap_by_distances(
        adata, distances, k=umap_k, dimensions=dimensions, min_dist=min_dist, spread=spread, random_seed=random_seed
    )


def _build_igraph(edge_weights: ut.Matrix) -> Tuple[ig.Graph, ut.NumpyVector]:
    edge_weights = ut.to_proper_matrix(edge_weights)
    assert edge_weights.shape[0] == edge_weights.shape[1]
    size = edge_weights.shape[0]

    sources, targets = edge_weights.nonzero()
    weights_array = ut.to_numpy_vector(edge_weights[sources, targets]).astype("float64")

    graph = ig.Graph(directed=True)
    graph.add_vertices(size)
    graph.add_edges(list(zip(sources, targets)))
    graph.es["weight"] = weights_array

    return graph, weights_array



"""
Parallel
--------

Due to the notorious GIL, using multiple Python threads is essentially useless. This leaves us with
two options for using multiple processors, which is mandatory for reasonable performance on the
large data sets we work on:

* Use multiple threads in the internal C++ implementation of some Python functions; this is done by both numpy and the
  C++ extension functions provided by this package, and works even for reasonably small sized work, such as sorting each
  of the rows of a large matrix.

* Use Python multi-processing. This is costly and works only for large sized work, such as computing metacells for
  different piles.

Each of these two approaches works tolerably well on its own, even though both are sub-optimal. The
problem starts when we want to combine them. Consider a server with 50 processors. Invoking
``corrcoef`` on a large matrix will use them all. This is great if one computes metacells for a
single pile. Suppose, however, you want to compute metacells for 50 piles, and do so using
multi-processing. Each and every of the 50 sub-processes will invoke ``corcoeff`` which will spawn
50 internal threads, resulting in the operating system seeing 2500 processes competing for the same
50 hardware processors. "This does not end well."

You would expect that, two decades after multi-core systems became available, this would have been
solved "out of the box" by the parallel frameworks (Python, OpenMP, TBB, etc.) all agreeing to
cooperate with each other. However, somehow this isn't seen as important by the people maintaining
these frameworks; in fact, most of them don't properly handle nested parallelism within their own
framework, never mind playing well with others.

So in practice, while languages built for parallelism (such as Julia and Rust) deal well with nested
parallel construct, using a mixture of older serial languages (such as Python and C++) puts us in a
swamp, and "you can't build a castel in a swamp". In our case, numpy uses some underlying parallel
threads framework, our own extensions uses OpenMP parallel threads, and we are forced to use the
Python-multi-processing framework itself on top of both, and each of these frameworks is blind to
the others.

As a crude band-aid, we force both whatever-numpy-uses and OpenMP to use a specific number of
threads. So, when we use multi-processing, we limit each sub-process to use less internal threads,
such that the total will be at most 50. This very sub-optimal, but at least it doesn't bring the
server to its knees trying to deal with a total load of 2500 processes.

A final twist on all this is that hyper-threading is (worse than) useless for heavy compute threads.
We therefore by default only use one thread per physical cores. We get the number pf physical cores
using the ``psutil`` package.

.. todo::

    Re-implement all the package in a single language more suitable for scientific computing. Julia
    is looking like a good combination of convenience and performance...
"""
import ctypes
import os
import sys
from math import ceil
from multiprocessing import Value
from multiprocessing import get_context
from threading import current_thread
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar

import psutil  # type: ignore
from threadpoolctl import threadpool_limits  # type: ignore

import metacells.utilities.documentation as utd
import metacells.utilities.logging as utl
import metacells.utilities.progress as utp
import metacells.utilities.timing as utm

if "sphinx" not in sys.argv[0]:
    import metacells.extensions as xt  # type: ignore

__all__ = [
    "is_main_process",
    "set_processors_count",
    "get_processors_count",
    "parallel_map",
]


PROCESSORS_COUNT = 0

MAIN_PROCESS_PID = os.getpid()

IS_MAIN_PROCESS: Optional[bool] = True

MAP_INDEX = 0
PROCESS_INDEX = 0

PROCESSES_COUNT = 0
NEXT_PROCESS_INDEX = Value(ctypes.c_int32, lock=True)
PARALLEL_FUNCTION: Optional[Callable[[int], Any]] = None


def is_main_process() -> bool:
    """
    Return whether this is the main process, as opposed to a sub-process spawned by
    :py:func:`parallel_map`.
    """
    return bool(IS_MAIN_PROCESS)


def set_processors_count(processors: int) -> None:
    """
    Set the (maximal) number of processors to use in parallel.

    The default value of ``0`` means using all the available physical processors. Note that if
    hyper-threading is enabled, this would be less than (typically half of) the number of logical
    processors in the system. This is intentional, as there's no value - actually, negative
    value - in running multiple heavy computations on hyper-threads of the same physical processor.

    Otherwise, the value is the actual (positive) number of processors to use. Override this by
    setting the ``METACELLS_PROCESSORS_COUNT`` environment variable or by invoking this function
    from the main thread.
    """
    assert IS_MAIN_PROCESS

    if processors == 0:
        processors = psutil.cpu_count(logical=False)

    assert processors > 0

    global PROCESSORS_COUNT
    PROCESSORS_COUNT = processors

    threadpool_limits(limits=PROCESSORS_COUNT)
    xt.set_threads_count(PROCESSORS_COUNT)
    os.environ["OMP_NUM_THREADS"] = str(PROCESSORS_COUNT)
    os.environ["MKL_NUM_THREADS"] = str(PROCESSORS_COUNT)


if "sphinx" not in sys.argv[0]:
    set_processors_count(int(os.environ.get("METACELLS_PROCESSORS_COUNT", "0")))


def get_processors_count() -> int:
    """
    Return the number of PROCESSORs we are allowed to use.
    """
    assert PROCESSORS_COUNT > 0
    return PROCESSORS_COUNT


T = TypeVar("T")


@utd.expand_doc()
def parallel_map(
    function: Callable[[int], T],
    invocations: int,
    *,
    max_processors: int = 0,
    hide_from_progress_bar: bool = False,
) -> List[T]:
    """
    Execute ``function``, in parallel, ``invocations`` times. Each invocation is given the invocation's index as its
    single argument.

    For our simple pipelines, only the main process is allowed to execute functions in parallel processes, that is, we
    do not support nested ``parallel_map`` calls.

    This uses :py:func:`get_processors_count` processes. If ``max_processors`` (default: {max_processors}) is zero, use
    all available processors. Otherwise, further reduces the number of processes used to at most the specified value.

    If this ends up using a single process, runs the function serially. Otherwise, fork new processes to execute the
    function invocations (using ``multiprocessing.get_context('fork').Pool.map``).

    The downside is that this is slow, and you need to set up **mutable** shared memory (e.g. for large results) in
    advance. The upside is that each of these processes starts with a shared memory copy(-on-write) of the full Python
    state, that is, all the inputs for the function are available "for free".

    If a progress bar is active at the time of invoking ``parallel_map``, and ``hide_from_progress_bar`` is not set,
    then it is assumed the parallel map will cover all the current (slice of) the progress bar, and it is reported into
    it in increments of ``1/invocations``.

    .. todo::

        It is currently only possible to invoke :py:func:`parallel_map` from the main application thread (that is, it
        does not nest).
    """
    if invocations == 0:
        return []

    assert function.__is_timed__  # type: ignore

    global IS_MAIN_PROCESS
    assert IS_MAIN_PROCESS

    global PROCESSES_COUNT
    PROCESSES_COUNT = min(PROCESSORS_COUNT, invocations)
    if max_processors != 0:
        assert max_processors > 0
        PROCESSES_COUNT = min(PROCESSES_COUNT, max_processors)

    if PROCESSES_COUNT == 1:
        return [function(index) for index in range(invocations)]

    NEXT_PROCESS_INDEX.value = 0  # type: ignore

    global PARALLEL_FUNCTION
    assert PARALLEL_FUNCTION is None

    global MAP_INDEX
    MAP_INDEX += 1

    num_threads = str(ceil(PROCESSES_COUNT / invocations))
    os.environ["OMP_NUM_THREADS"] = num_threads
    os.environ["MKL_NUM_THREADS"] = num_threads

    PARALLEL_FUNCTION = function
    IS_MAIN_PROCESS = None
    try:
        results: List[Optional[T]] = [None] * invocations
        utm.flush_timing()
        with utm.timed_step("parallel_map"):
            utm.timed_parameters(index=MAP_INDEX, processes=PROCESSES_COUNT)
            with get_context("fork").Pool(PROCESSES_COUNT) as pool:
                for index, result in pool.imap_unordered(_invocation, range(invocations)):
                    if utp.has_progress_bar() and not hide_from_progress_bar:
                        utp.did_progress(1 / invocations)
                    results[index] = result
        return results  # type: ignore
    finally:
        IS_MAIN_PROCESS = True
        PARALLEL_FUNCTION = None
        os.environ["OMP_NUM_THREADS"] = str(PROCESSES_COUNT)
        os.environ["MKL_NUM_THREADS"] = str(PROCESSES_COUNT)


def _invocation(index: int) -> Tuple[int, Any]:
    global IS_MAIN_PROCESS
    if IS_MAIN_PROCESS is None:
        IS_MAIN_PROCESS = os.getpid() == MAIN_PROCESS_PID
        assert not IS_MAIN_PROCESS

        global PROCESS_INDEX
        with NEXT_PROCESS_INDEX:
            PROCESS_INDEX = NEXT_PROCESS_INDEX.value  # type: ignore
            NEXT_PROCESS_INDEX.value += 1  # type: ignore

        current_thread().name = f"#{MAP_INDEX}.{PROCESS_INDEX}"
        utm.in_parallel_map(MAP_INDEX, PROCESS_INDEX)

        global PROCESSORS_COUNT
        start_processor_index = int(round(PROCESSORS_COUNT * PROCESS_INDEX / PROCESSES_COUNT))
        stop_processor_index = int(round(PROCESSORS_COUNT * (PROCESS_INDEX + 1) / PROCESSES_COUNT))
        PROCESSORS_COUNT = stop_processor_index - start_processor_index

        assert PROCESSORS_COUNT > 0
        utl.logger().debug("PROCESSORS: %s", PROCESSORS_COUNT)
        threadpool_limits(limits=PROCESSORS_COUNT)
        xt.set_threads_count(PROCESSORS_COUNT)
        os.environ["OMP_NUM_THREADS"] = str(PROCESSORS_COUNT)
        os.environ["MKL_NUM_THREADS"] = str(PROCESSORS_COUNT)

    assert PARALLEL_FUNCTION is not None
    result = PARALLEL_FUNCTION(index)
    return index, result



"""
Default metacell processing pipeline.

The functions here are thin wrappers which invoke a series of steps (:py:mod:`metacells.tools`) to
provide a complete pipeline for computing metacells for your data.

This pipeline can be configured by tweaking the (many) parameters, but for any deeper customization
(adding and/or removing steps) just provide your own pipeline instead. You can use the
implementation here as a starting point.

All the functions included here are exported under ``metacells.pl``.
"""

from .collect import *
from .direct import *
from .divide_and_conquer import *
from .exclude import *
from .mark import *
from .mcview import *
from .projection import *
from .related_genes import *
from .select import *
from .umap import *



"""
Layout
------
"""

import warnings
from typing import Optional
from typing import Union

import numpy as np
import scipy.sparse as sp  # type: ignore
import umap  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "umap_by_distances",
    "spread_coordinates",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def umap_by_distances(
    adata: AnnData,
    distances: Union[str, ut.ProperMatrix] = "umap_distances",
    *,
    prefix: str = "",
    k: int = pr.umap_k,
    dimensions: int = 2,
    min_dist: float = pr.umap_min_dist,
    spread: float = pr.umap_spread,
    random_seed: int,
) -> None:
    """
    Compute layout for the observations using UMAP, based on a distances matrix.

    **Input**

    The input annotated ``adata`` is expected to contain a per-observation-per-observation property
    ``distances`` (default: {distances}), which describes the distance between each two observations
    (cells). The distances must be non-negative, symmetrical, and zero for self-distances (on the
    diagonal).

    **Returns**

    Sets the following annotations in ``adata``:

    Observation (Cell) Annotations
        ``<prefix>x``, ``<prefix>y``
            Coordinates for UMAP 2D projection of the observations (if ``dimensions`` is 2).
        ``<prefix>u``, ``<prefix>v``, ``<prefix>w``
            Coordinates for UMAP 3D projection of the observations (if ``dimensions`` is 3).

    **Computation Parameters**

    1. Invoke UMAP to compute a layout of some ``dimensions`` (default: {dimensions}D) using
       ``min_dist`` (default: {min_dist}), ``spread`` (default: {spread}) and ``k`` (default: {k}).
       If the spread is lower than the minimal distance, it is raised. If ``random_seed``
       is not zero, then it is passed to UMAP to force the computation to be
       reproducible. However, this means UMAP will use a single-threaded implementation that will be
       slower.
    """
    assert dimensions in (2, 3)
    if isinstance(distances, str):
        distances_matrix = ut.get_oo_proper(adata, distances)
    else:
        distances_matrix = distances
    # UMAP dies when given a dense matrix.
    distances_csr = sp.csr_matrix(distances_matrix)

    spread = max(min_dist, spread)  # UMAP insists.

    # UMAP implementation doesn't know to reduce K by itself.
    n_neighbors = min(k, adata.n_obs - 2)

    random_state: Optional[int] = None
    if random_seed != 0:
        random_state = random_seed

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            coordinates = umap.UMAP(
                metric="precomputed",
                n_neighbors=n_neighbors,
                spread=spread,
                min_dist=min_dist,
                n_components=dimensions,
                random_state=random_state,
            ).fit_transform(distances_csr)
    except ValueError:
        # UMAP implementation doesn't know how to handle too few edges.
        # However, it considers structural zeros as real edges.
        distances_matrix = ut.to_numpy_matrix(distances_matrix, copy=True)
        distances_matrix += 1.0
        np.fill_diagonal(distances_matrix, 0.0)
        distances_csr = sp.csr_matrix(distances_matrix)
        distances_csr.data -= 1.0
        coordinates = umap.UMAP(
            metric="precomputed", n_neighbors=n_neighbors, spread=spread, min_dist=min_dist, random_state=random_state
        ).fit_transform(distances_csr)

    all_sizes = []
    all_coordinates = []
    for axis in range(dimensions):
        axis_coordinates = ut.to_numpy_vector(coordinates[:, axis], copy=True)
        min_coordinate = np.min(coordinates)
        max_coordinate = np.max(coordinates)
        size = max_coordinate - min_coordinate
        assert size > 0
        all_sizes.append(size)
        all_coordinates.append(axis_coordinates)

    if dimensions == 2:
        all_names = ["x", "y"]
    elif dimensions == 3:
        all_names = ["u", "v", "w"]
    else:
        assert False

    order = np.argsort(all_sizes)
    for axis, name in zip(order, all_names):
        ut.set_o_data(adata, f"{prefix}{name}", all_coordinates[axis])


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def spread_coordinates(
    adata: AnnData,
    *,
    prefix: str = "",
    suffix: str = "_spread",
    cover_fraction: float = pr.spread_cover_fraction,
    noise_fraction: float = pr.spread_noise_fraction,
    random_seed: int,
) -> None:
    """
    Move UMAP points so they cover some fraction of the plot area without overlapping.

    **Input**

    The input annotated ``adata`` is expected to contain the per-observation properties
    ``<prefix>x`` and ``<prefix>y`` (default prefix: {prefix}) which contain the UMAP coordinates.

    **Returns**

    Sets the following annotations in ``adata``:

    Observation (Cell) Annotations
        ``<prefix>x<suffix>``, ``<prefix>y<suffix>`` (default suffix: {suffix})
            The new coordinates which will be spread out so the points do not overlap and
            cover some fraction of the total plot area.

    **Computation Parameters**

    1. Move the points so they cover ``cover_fraction`` (default: {cover_fraction}) of the total
       plot area. Also add a noise of the ``noise_fraction`` (default: {noise_fraction}) of the
       minimal distance between the points. A non-zero ``random_seed`` will make this reproducible.
    """
    assert 0 < cover_fraction < 1
    assert noise_fraction >= 0

    x_coordinates = ut.get_o_numpy(adata, f"{prefix}x")
    y_coordinates = ut.get_o_numpy(adata, f"{prefix}y")

    x_coordinates, y_coordinates = ut.cover_coordinates(
        x_coordinates,
        y_coordinates,
        cover_fraction=cover_fraction,
        noise_fraction=noise_fraction,
        random_seed=random_seed,
    )

    ut.set_o_data(adata, f"{prefix}x{suffix}", x_coordinates)
    ut.set_o_data(adata, f"{prefix}y{suffix}", y_coordinates)



"""
Filter
------
"""
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.utilities as ut
from metacells.tools.mask import combine_masks

__all__ = [
    "filter_data",
]


@ut.logged()
@ut.timed_call()
def filter_data(  # pylint: disable=dangerous-default-value
    adata: AnnData,
    obs_masks: List[str] = [],
    var_masks: List[str] = [],
    *,
    mask_obs: Optional[str] = None,
    mask_var: Optional[str] = None,
    invert_obs: bool = False,
    invert_var: bool = False,
    track_obs: Optional[str] = None,
    track_var: Optional[str] = None,
    name: Optional[str] = None,
    top_level: bool = True,
) -> Optional[Tuple[AnnData, ut.PandasSeries, ut.PandasSeries]]:
    """
    Filter (slice) the data based on previously-computed masks.

    For example, it is useful to discard cell-cycle genes, cells which have too few UMIs for
    meaningful analysis, etc. In general, the "best" filter depends on the data set.

    This function makes it easy to combine different pre-computed per-observation (cell) and
    per-variable (gene) boolean mask annotations into a final overall inclusion mask, and slice the
    data accordingly, while tracking the base index of the cells and genes in the filtered data.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    An annotated data containing a subset of the observations (cells) and variables (genes).

    If no observations and/or no variables were selected by the filter, returns ``None``.

    If ``name`` is not specified, the returned data will be unnamed. Otherwise, if the name starts
    with a ``.``, it will be appended to the current name (if any). Otherwise, ``name`` is the new
    name.

    If ``mask_obs`` and/or ``mask_var`` are specified, store the mask of the selected data as a
    per-observation and/or per-variable annotation of the full ``adata``.

    If ``track_obs`` and/or ``track_var`` are specified, store the original indices of the selected
    data as a per-observation and/or per-variable annotation of the result data.

    **Computation Parameters**

    1. Combine the masks in ``obs_masks`` and/or ``var_masks`` using
       :py:func:`metacells.tools.mask.combine_masks` passing it ``invert_obs`` and ``invert_var``,
       and ``mask_obs`` and ``mask_var`` as the ``to`` parameter. If either list of masks is empty,
       use the full mask.

    2. If the obtained masks for either the observations or variables is empty, return ``None``.
       Otherwise, return a slice of the full data containing just the observations and variables
       specified by the final masks.
    """
    if len(obs_masks) == 0:
        obs_mask = np.full(adata.n_obs, True, dtype="bool")
        if mask_obs is not None:
            ut.set_o_data(adata, mask_obs, obs_mask)
    else:
        mask = combine_masks(adata, obs_masks, invert=invert_obs, to=mask_obs)
        if mask is None:
            assert mask_obs is not None
            obs_mask = ut.get_o_numpy(adata, mask_obs, formatter=ut.mask_description) > 0
        else:
            obs_mask = ut.to_numpy_vector(mask, only_extract=True) > 0

    if len(var_masks) == 0:
        var_mask = np.full(adata.n_vars, True, dtype="bool")
        if mask_var is not None:
            ut.set_o_data(adata, mask_var, var_mask)
    else:
        mask = combine_masks(adata, var_masks, invert=invert_var, to=mask_var)
        if mask is None:
            assert mask_var is not None
            var_mask = ut.get_v_numpy(adata, mask_var, formatter=ut.mask_description) > 0
        else:
            var_mask = ut.to_numpy_vector(mask, only_extract=True) > 0

    if not np.any(obs_mask) or not np.any(var_mask):
        return None

    fdata = ut.slice(
        adata, name=name, top_level=top_level, obs=obs_mask, vars=var_mask, track_obs=track_obs, track_var=track_var
    )

    return (
        fdata,
        ut.to_pandas_series(obs_mask, index=adata.obs_names),
        ut.to_pandas_series(var_mask, index=adata.var_names),
    )



"""
Collect
-------
"""

from hashlib import shake_128
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import scipy.sparse as sp  # type: ignore
import scipy.stats as ss  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut
from metacells import __version__  # pylint: disable=cyclic-import

__all__ = [
    "collect_metacells",
]


@ut.logged()
@ut.timed_call()
def collect_metacells(  # pylint: disable=too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    metacell_geo_mean: bool = pr.metacell_geo_mean,
    metacell_umis_regularization: float = pr.metacell_umis_regularization,
    zeros_cell_size_quantile: float = pr.zeros_cell_size_quantile,
    groups: Union[str, ut.Vector] = "metacell",
    name: str = "metacells",
    prefix: Optional[str] = "M",
    top_level: bool = True,
    _metacell_groups: bool = False,
    random_seed: int,
) -> AnnData:
    """
    Collect computed metacells ``what`` (default: {what}) data.

    **Input**

    Annotated (presumably "clean") ``adata``, where the observations are cells and the variables are genes, and where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation
    containing such a matrix.

    **Returns**

    Annotated metacell data containing for each observation (metacell) for each variable (gene) the fraction of the gene
    in the metacell, and the following annotations:

    Unstructured (Scalar)
        ``outliers``
            The number of outlier cells (which were not assigned to any metacell).

        ``metacells_algorithm``
            The version of the algorithm generating the metacell (based on the package version number).

    Variable (Gene) Annotations
        ``*``
            Every per-gene annotations of ``adata`` is copied to the result.

    Observations (Metacell) Annotations
        ``grouped``
            The number of ("clean") cells grouped into each metacell.

        ``total_umis``
            The total of all the UMIs of all the genes of all the cells grouped into the metacell.

    Observations-Variables (Metacell-Gene) Annotations
        ``total_umis``
            The total of all the UMIs of each genes in all the cells grouped into the metacell.

    Also sets in the full ``adata``:

    Observations (Cell) Annotations
        ``metacell_name``
            The string name of the metacell, which is ``prefix`` (default: {prefix}) followed by the metacell index,
            followed by ``.checksum`` where the checksum is a 2 digits, reflecting the (names of the) set of cells
            grouped into each metacell. This keeps metacell names short, and provides protection against assigning
            per-metacell annotations from a different metacell computation to the new results (unless they are
            identical).

    **Computation Parameters**

    For each metacell:

    3. If ``metacell_geo_mean`` (default: {metacell_geo_mean}),

        A. Compute the fraction of each gene out of each cell grouped into the metacell.

        B. For each cell, add to the fractions the ``metacell_umis_regularization`` divided by the total UMIs of the
           cell.

        C. For each gene, compute the weighted geomean of these fractions across all the cells, where the weight of each
           cell is the log of its total number of UMIs, and

        D. Subtract the geomean of the per-cell regularization so all-zero genes would have a zero fraction.

    4. Otherwise, for each metacell, sum the total UMIs of each gene across all cells, and divide it by the total UMIs
       of all genes in all cells.

    5. Normalize the per-gene fractions so their sum would be 1.0 in the metacell.

    6. Compute the ``zeros_cell_size_quantile`` of the total UMIs of the cells to pick the number of UMIs to use for
       zeros computation.

    7. For each gene in each cell, compute the probability it would be zero if we downsampled the cell to have this
       number of UMIs.

    8. Use these probabilities to decide whether the each gene is "effectively zero" in each cell.

    9. For each gene, count the number of cells in which it is "effectively zero".

    In addition:

    10. Pass all relevant per-gene annotations from ``adata`` to the result.

    11. Set the ``metacell_name`` property of each cell in ``adata`` to the name of the group (metacell) it belongs to.
        Cells which do not belong to any metacell are assigned the metacell name ``Outliers``.
    """
    metacell_of_cells = ut.get_o_numpy(adata, groups, formatter=ut.groups_description).copy()
    metacell_of_cells[metacell_of_cells < 0] = -1
    outliers_count = np.sum(metacell_of_cells < 0)
    metacells_count = np.max(metacell_of_cells) + 1
    ut.log_calc("outliers_count", outliers_count)
    ut.log_calc("metacells_count", metacells_count)

    umis_per_gene_per_cell = ut.get_vo_proper(adata, what, layout="row_major")
    umis_per_cell = ut.get_o_numpy(adata, what, sum=True).astype("float64")

    @ut.timed_call()
    def _collect_metacell(metacell_index: int) -> Dict[str, Any]:
        with ut.log_step(
            "- metacell",
            metacell_index,
            formatter=lambda metacell_index: ut.progress_description(metacells_count, metacell_index, "metacell"),
        ):
            if random_seed != 0:
                np.random.seed(random_seed + metacell_index)

            mask_per_cell = metacell_of_cells == metacell_index
            grouped_of_metacell = np.sum(mask_per_cell)
            ut.log_calc("grouped_of_metacell", grouped_of_metacell)
            assert grouped_of_metacell > 0

            umis_per_cell_of_metacell = umis_per_cell[mask_per_cell]
            total_umis_of_metacell = np.sum(umis_per_cell_of_metacell.astype("int64"))
            ut.log_calc("total_umis_of_metacell", total_umis_of_metacell)

            umis_per_gene_per_cell_of_metacell = ut.to_numpy_matrix(umis_per_gene_per_cell[mask_per_cell, :])

            zeros_downsample_umis = round(np.quantile(umis_per_cell_of_metacell, zeros_cell_size_quantile))
            ut.log_calc("zeros_downsample_umis", zeros_downsample_umis)

            downsampled_umis_per_gene_per_cell_of_metacell = ut.downsample_matrix(
                umis_per_gene_per_cell_of_metacell, per="row", samples=zeros_downsample_umis, random_seed=random_seed
            )
            zeros_per_gene_per_cell_of_metacell = downsampled_umis_per_gene_per_cell_of_metacell == 0
            zeros_per_gene_per_cell_of_metacell = ut.to_layout(
                zeros_per_gene_per_cell_of_metacell, layout="column_major"  # type: ignore
            )

            zeros_per_gene = ut.sum_per(zeros_per_gene_per_cell_of_metacell, per="column").astype("int32")
            ut.log_calc("zeros_per_gene", zeros_per_gene, formatter=ut.sizes_description)

            umis_per_gene_per_cell_of_metacell = ut.to_layout(umis_per_gene_per_cell_of_metacell, layout="column_major")
            umis_per_gene_of_metacell = ut.sum_per(umis_per_gene_per_cell_of_metacell, per="column").astype("float32")

            if not metacell_geo_mean:
                total_downsampled_umis_of_metacell = np.sum(
                    downsampled_umis_per_gene_per_cell_of_metacell  # type: ignore
                )

                downsampled_umis_per_gene_per_cell_of_metacell = ut.to_layout(
                    downsampled_umis_per_gene_per_cell_of_metacell, layout="column_major"
                )
                downsampled_umis_per_gene_of_metacell = ut.sum_per(
                    downsampled_umis_per_gene_per_cell_of_metacell, per="column"
                ).astype("float32")
                assert np.sum(downsampled_umis_per_gene_of_metacell) == total_downsampled_umis_of_metacell

                fraction_per_gene_of_metacell = (
                    downsampled_umis_per_gene_of_metacell / total_downsampled_umis_of_metacell
                )

            else:
                fraction_per_gene_per_cell_of_metacell = ut.to_layout(
                    umis_per_gene_per_cell_of_metacell / ut.to_numpy_matrix(umis_per_cell_of_metacell[:, np.newaxis]),
                    layout="column_major",
                )

                regularization_per_cell_of_metacell = metacell_umis_regularization / umis_per_cell_of_metacell
                regularization_of_metacell = ss.mstats.gmean(regularization_per_cell_of_metacell)

                log_fraction_per_gene_per_cell_of_metacell = np.log2(
                    fraction_per_gene_per_cell_of_metacell + regularization_per_cell_of_metacell[:, np.newaxis]
                )
                weight_per_cell_of_metacell = np.log2(umis_per_cell_of_metacell)
                weighted_log_fraction_per_gene_per_cell_of_metacell = ut.to_layout(
                    log_fraction_per_gene_per_cell_of_metacell * weight_per_cell_of_metacell[:, np.newaxis],
                    layout="column_major",
                )

                total_weighted_log_fraction_per_gene_of_metacell = ut.sum_per(
                    weighted_log_fraction_per_gene_per_cell_of_metacell, per="column"
                )
                total_weight_of_metacell = np.sum(weight_per_cell_of_metacell)

                mean_log_fraction_per_gene_of_metacell = (
                    total_weighted_log_fraction_per_gene_of_metacell / total_weight_of_metacell
                )
                fraction_per_gene_of_metacell = 2**mean_log_fraction_per_gene_of_metacell
                fraction_per_gene_of_metacell -= regularization_of_metacell
                fraction_per_gene_of_metacell[umis_per_gene_of_metacell == 0] = 0
                fraction_per_gene_of_metacell[fraction_per_gene_of_metacell < 0] = 0
                assert np.min(fraction_per_gene_of_metacell) >= 0
                fraction_per_gene_of_metacell /= np.sum(fraction_per_gene_of_metacell)

            fraction_per_gene_of_metacell = fraction_per_gene_of_metacell.astype("float32")
            if _metacell_groups:
                fraction_per_gene_of_metacell *= total_umis_of_metacell
                ut.log_calc(
                    "effective_umis_per_gene_of_metacell", fraction_per_gene_of_metacell, formatter=ut.sizes_description
                )
            else:
                ut.log_calc(
                    "fraction_per_gene_of_metacell", fraction_per_gene_of_metacell, formatter=ut.sizes_description
                )

            return {
                "grouped": grouped_of_metacell,
                "total_umis": total_umis_of_metacell,
                "umis_per_gene": umis_per_gene_of_metacell,
                "fraction_per_gene": fraction_per_gene_of_metacell,
                "zeros_downsample_umis": zeros_downsample_umis,
                "zeros_per_gene": zeros_per_gene,
            }

    results = ut.parallel_map(_collect_metacell, metacells_count)

    fraction_per_gene_per_metacell = sp.csr_matrix(np.vstack([result["fraction_per_gene"] for result in results]))
    assert str(fraction_per_gene_per_metacell.dtype) == "float32"
    assert fraction_per_gene_per_metacell.shape == (metacells_count, adata.n_vars)

    mdata = AnnData(fraction_per_gene_per_metacell)
    if top_level:
        ut.top_level(mdata)
    if name.startswith("."):
        name = ut.get_name(adata, "unnamed") + name  # type: ignore
    ut.set_name(mdata, name)
    mdata.var_names = adata.var_names
    mdata.obs_names = _obs_names(prefix or "", ut.to_numpy_vector(adata.obs_names), metacell_of_cells)

    grouped_per_metacell = np.array([result["grouped"] for result in results])
    ut.set_o_data(mdata, "grouped", grouped_per_metacell)

    total_umis_per_metacell = np.array([result["total_umis"] for result in results])
    ut.set_o_data(mdata, "total_umis", total_umis_per_metacell)

    umis_per_gene_per_metacell = np.vstack([result["umis_per_gene"] for result in results])
    assert str(umis_per_gene_per_metacell.dtype) == "float32"
    assert umis_per_gene_per_metacell.shape == (metacells_count, adata.n_vars)
    ut.set_vo_data(mdata, "total_umis", umis_per_gene_per_metacell)

    zeros_downsample_umis_per_metacell = np.array([result["zeros_downsample_umis"] for result in results])
    ut.set_o_data(mdata, "__zeros_downsample_umis", zeros_downsample_umis_per_metacell)

    zeros_per_gene_per_metacell = np.vstack([result["zeros_per_gene"] for result in results])
    assert str(zeros_per_gene_per_metacell.dtype) == "int32"
    assert zeros_per_gene_per_metacell.shape == (metacells_count, adata.n_vars)
    ut.set_vo_data(mdata, "zeros", zeros_per_gene_per_metacell)

    metacell_names = np.array(["Outliers"] + list(mdata.obs_names), dtype="str")
    metacell_of_cells += 1
    metacell_name_of_cells = metacell_names[metacell_of_cells]
    ut.set_o_data(adata, "metacell_name", metacell_name_of_cells)

    for annotation_name in adata.var.keys():
        value_per_gene = ut.get_v_numpy(adata, annotation_name)
        ut.set_v_data(mdata, annotation_name, value_per_gene)

    if isinstance(groups, str) and ut.has_data(adata, "metacells_level"):
        tl.convey_obs_to_group(adata=adata, gdata=mdata, group=groups, property_name="metacell_level")
    if isinstance(groups, str) and ut.has_data(adata, "rare_cell"):
        tl.convey_obs_to_group(
            adata=adata,
            gdata=mdata,
            group=groups,
            property_name="cells_rare_gene_module",
            to_property_name="metacells_rare_gene_module",
        )
        tl.convey_obs_to_group(
            adata=adata,
            gdata=mdata,
            group=groups,
            property_name="rare_cell",
            to_property_name="rare_metacell",
        )

    ut.set_m_data(mdata, "outliers", outliers_count)
    ut.set_m_data(mdata, "metacells_algorithm", f"metacells.{__version__}")

    return mdata


# TODO: Replicated in metacells.tools.group
def _obs_names(prefix: str, name_of_members: ut.NumpyVector, group_of_members: ut.NumpyVector) -> List[str]:
    groups_count = np.max(group_of_members) + 1
    name_of_groups: List[str] = []
    prefix = prefix or ""
    for group_index in range(groups_count):
        groups_mask = group_of_members == group_index
        assert np.any(groups_mask)
        hasher = shake_128()
        for member_name in name_of_members[groups_mask]:
            hasher.update(member_name.encode("utf8"))
        checksum = int(hasher.hexdigest(16), 16) % 100
        name_of_groups.append(f"{prefix}{group_index}.{checksum:02d}")
    return name_of_groups



"""
Timing
------

The first step in achieving reasonable performance is identifying where most of the time is being
spent. The functions in this module allow to easily collect timing information about the relevant
functions or steps within functions in a controlled way, with low overhead, as opposed to collecting
information about all functions which has higher overheads and produces mountains of mostly
irrelevant data.
"""

import os
from contextlib import contextmanager
from functools import wraps
from threading import current_thread
from threading import local as thread_local
from time import perf_counter_ns
from time import process_time_ns
from typing import IO
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import TypeVar

import metacells.utilities.documentation as utd
import metacells.utilities.logging as utl

__all__ = [
    "collect_timing",
    "flush_timing",
    "in_parallel_map",
    "log_steps",
    "timed_step",
    "timed_call",
    "timed_parameters",
    "context",
    "current_step",
    "StepTiming",
    "Counters",
]

COLLECT_TIMING = False

TIMING_PATH = "timing.csv"
TIMING_MODE = "a"
TIMING_BUFFERING = 1
TIMING_FILE: Optional[IO] = None

LOG_ALL_STEPS = False

THREAD_LOCAL = thread_local()
COUNTED_THREADS = 0


@utd.expand_doc()
def collect_timing(
    collect: bool,
    path: str = TIMING_PATH,  # pylint: disable=used-prior-global-declaration
    mode: str = TIMING_MODE,  # pylint: disable=used-prior-global-declaration
    *,
    buffering: int = TIMING_BUFFERING,  # pylint: disable=used-prior-global-declaration
) -> None:
    """
    Specify whether, where and how to collect timing information.

    By default, we do not. Override this by setting the ``METACELLS_COLLECT_TIMING`` environment
    variable to ``true``, or by invoking this function from the main thread.

    By default, the data is written to the ``path`` is {path}, which is opened with the mode is
    {mode} and using the buffering is {buffering}. Override this by setting the
    ``METACELL_TIMING_PATH``, ``METACELL_TIMING_MODE`` and/or the ``METACELL_TIMING_BUFFERING``
    environment variables, or by invoking this function from the main thread.

    This will flush and close the previous timing file, if any.

    The file is written in CSV format (without headers). The first three fields are:

    * The invocation context (a ``.``-separated path of "relevant" function/step names).

    * The elapsed time (in nanoseconds) in this context (not counting nested contexts).

    * The CPU time (in nanoseconds) in this context (not counting nested contexts).

    This may be followed by a series of ``name,value`` pairs describing parameters of interest for
    this context, such as data sizes and layouts, to help understand the performance of the code.
    """
    global TIMING_PATH
    global TIMING_MODE
    global TIMING_BUFFERING
    global TIMING_FILE
    global COLLECT_TIMING

    if not path.endswith(".csv"):
        raise ValueError(f"The METACELL_TIMING_PATH: {path} does not end with: .csv")

    TIMING_PATH = path
    TIMING_MODE = mode
    TIMING_BUFFERING = buffering

    if TIMING_FILE is not None:
        TIMING_FILE.flush()
        TIMING_FILE.close()
        TIMING_FILE = None

    if collect:
        TIMING_FILE = open(  # pylint: disable=consider-using-with
            TIMING_PATH, TIMING_MODE, buffering=TIMING_BUFFERING, encoding="utf8"
        )

    COLLECT_TIMING = collect


def flush_timing() -> None:
    """
    Flush the timing information, if we are collecting it.
    """
    if TIMING_FILE is not None:
        TIMING_FILE.flush()


def in_parallel_map(map_index: int, process_index: int) -> None:
    """
    Reconfigure timing collection when running in a parallel sub-process via
    :py:func:`metacells.utilities.parallel.parallel_map`.

    This will direct the timing information from ``<timing>.csv`` to
    ``<timing>.<map>.<process>.csv`` (where ``<timing>`` is from the original path, ``<map>`` is the
    serial number of the :py:func:`metacells.utilities.parallel.parallel_map` invocation, and
    ``<process>`` is the serial number of the process in the map).

    Collecting the timing of separate sub-processes to separate files allows us to freely write to
    them without locks and synchronizations which improves the performance (reduces the overhead of
    collecting timing information).

    You can just concatenate the files when the run is complete, or use a tool which automatically
    collects the data from all the files, such as :py:mod:`metacells.scripts.timing`.
    """
    if COLLECT_TIMING:
        assert TIMING_PATH.endswith(".csv")
        collect_timing(True, f"{TIMING_PATH[:-4]}.{map_index}.{process_index}.csv")


def log_steps(log: bool) -> None:
    """
    Whether to log every step invocation.

    By default, we do not. Override this by setting the ``METACELLS_LOG_ALL_STEPS`` environment
    variable to ``true`` or by invoking this function from the main thread.

    .. note::

        This only works if :py:func:`collect_timing` was set. It is a crude instrument to hunt for
        deadlocks, very-long-running numpy functions, and the like. Basically, if the program is
        taking 100% CPU and you have no idea what it is doing, turning this on and looking at the
        last logged step name would give you some idea of where it is stuck.
    """
    global LOG_ALL_STEPS
    LOG_ALL_STEPS = log


TIMING_PATH = os.environ.get("METACELL_TIMING_CSV", TIMING_PATH)
TIMING_MODE = os.environ.get("METACELL_TIMING_MODE", TIMING_MODE)
TIMING_BUFFERING = int(os.environ.get("METACELL_TIMING_BUFFERING", str(TIMING_BUFFERING)))
collect_timing({"true": True, "false": False}[os.environ.get("METACELLS_COLLECT_TIMING", str(COLLECT_TIMING)).lower()])
log_steps({"true": True, "false": False}[os.environ.get("METACELLS_LOG_ALL_STEPS", str(LOG_ALL_STEPS)).lower()])


class Counters:
    """
    The counters for the execution times.
    """

    def __init__(self, *, elapsed_ns: int = 0, cpu_ns: int = 0) -> None:
        self.elapsed_ns = elapsed_ns  #: Elapsed time counter.
        self.cpu_ns = cpu_ns  #: CPU time counter.

    @staticmethod
    def now() -> "Counters":
        """
        Return the current value of the counters.
        """
        return Counters(elapsed_ns=perf_counter_ns(), cpu_ns=process_time_ns())

    def __add__(self, other: "Counters") -> "Counters":
        return Counters(elapsed_ns=self.elapsed_ns + other.elapsed_ns, cpu_ns=self.cpu_ns + other.cpu_ns)

    def __iadd__(self, other: "Counters") -> "Counters":
        self.elapsed_ns += other.elapsed_ns
        self.cpu_ns += other.cpu_ns
        return self

    def __sub__(self, other: "Counters") -> "Counters":
        return Counters(elapsed_ns=self.elapsed_ns - other.elapsed_ns, cpu_ns=self.cpu_ns - other.cpu_ns)

    def __isub__(self, other: "Counters") -> "Counters":
        self.elapsed_ns -= other.elapsed_ns
        self.cpu_ns -= other.cpu_ns
        return self


class StepTiming:
    """
    Timing information for some named processing step.
    """

    def __init__(self, name: str, parent: Optional["StepTiming"]) -> None:
        """
        Start collecting time for a named processing step.
        """
        #: The parent step, if any.
        self.parent = parent

        if name[0] != ".":
            name = ";" + name

        if parent is None:
            assert not name[0] == "."
            name = name[1:]

        #: The full context of the processing step.
        self.context: str = name if parent is None else parent.context + name

        #: Parameters of interest of the processing step.
        self.parameters: List[str] = []

        #: The thread the step was invoked in.
        self.thread_name = current_thread().name

        #: The amount of CPU used in nested steps in the same thread.
        self.total_nested = Counters()


class GcStep(NamedTuple):
    """
    Data about a GC collection step.
    """

    #: The counters when we started the GC step.
    start: Counters

    #: The counters when we ended the GC step.
    stop: Counters


if COLLECT_TIMING:
    import gc

    GC_STEPS: List[GcStep] = []

    GC_START_POINT: Optional[Counters] = None

    def _time_gc(phase: str, info: Dict[str, Any]) -> None:
        steps_stack = getattr(THREAD_LOCAL, "steps_stack", None)
        if not steps_stack:
            return

        global GC_START_POINT
        if phase == "start":
            assert GC_START_POINT is None
            GC_START_POINT = Counters.now()
            return

        assert phase == "stop"
        assert GC_START_POINT is not None

        gc_step = GcStep(start=GC_START_POINT, stop=Counters.now())
        GC_STEPS.append(gc_step)
        GC_START_POINT = None

        gc_parameters = []
        for name, value in info.items():
            gc_parameters.append(name)
            gc_parameters.append(str(value))

        _print_timing("__gc__", gc_step.stop - gc_step.start, gc_parameters)

    gc.callbacks.append(_time_gc)


@contextmanager
def timed_step(name: str) -> Iterator[None]:  # pylint: disable=too-many-branches
    """
    Collect timing information for a computation step.

    Expected usage is:

    .. code:: python

        with ut.timed_step("foo"):
            some_computation()

    If we are collecting timing information, then for every invocation, the program will append a
    line similar to:

    .. code:: text

        foo,elapsed_ns,123,cpu_ns,456

    To a timing log file (default: ``timing.csv``). Additional fields can be appended to the line
    using the ``metacells.utilities.timing.parameters`` function.

    If the ``name`` starts with a ``.`` of a ``_``, then it is prefixed with the names of the
    innermost surrounding step name (which must exist). This is commonly used to time sub-steps of a
    function.
    """
    if not COLLECT_TIMING:
        yield None
        return

    steps_stack = getattr(THREAD_LOCAL, "steps_stack", None)
    if steps_stack is None:
        steps_stack = THREAD_LOCAL.steps_stack = []

    parent_timing: Optional[StepTiming] = None
    if len(steps_stack) > 0:
        parent_timing = steps_stack[-1]
    if name[0] == "_":
        name = f".{name[1:]}"
    if name[0] == ".":
        assert parent_timing is not None

    step_timing = StepTiming(name, parent_timing)
    steps_stack.append(step_timing)

    if LOG_ALL_STEPS:
        utl.logger().debug(f"{{[( {step_timing.context}")  # pylint: disable=logging-fstring-interpolation

    yield_point = Counters.now()
    try:
        yield None

    finally:
        back_point = Counters.now()
        total_times = back_point - yield_point

        global GC_STEPS
        gc_steps: List[GcStep] = []
        for gc_step in GC_STEPS:
            if yield_point.elapsed_ns <= gc_step.start.elapsed_ns and gc_step.stop.elapsed_ns <= back_point.elapsed_ns:
                total_times -= gc_step.stop - gc_step.start
            else:
                gc_steps.append(gc_step)
        GC_STEPS = gc_steps

        assert total_times.elapsed_ns >= 0
        assert total_times.cpu_ns >= 0
        if LOG_ALL_STEPS:
            utl.logger().debug("}]) %s", step_timing.context)

        steps_stack.pop()

        if parent_timing is not None:
            parent_timing.total_nested += total_times

        total_times -= step_timing.total_nested

        _print_timing(step_timing.context, total_times, step_timing.parameters)

        assert total_times.elapsed_ns >= 0
        assert total_times.cpu_ns >= 0


def _print_timing(
    invocation_context: str,
    total_times: Counters,
    step_parameters: Optional[List[str]] = None,
) -> None:
    gc_enabled = gc.isenabled()
    gc.disable()

    try:
        global TIMING_FILE
        if TIMING_FILE is None:
            TIMING_FILE = open(  # pylint: disable=consider-using-with
                TIMING_PATH, "a", buffering=TIMING_BUFFERING, encoding="utf8"
            )
        text = [invocation_context, "elapsed_ns", str(total_times.elapsed_ns), "cpu_ns", str(total_times.cpu_ns)]
        if step_parameters:
            text.extend(step_parameters)
        TIMING_FILE.write(",".join(text) + "\n")

    finally:
        if gc_enabled:
            gc.enable()


def timed_parameters(**kwargs: Any) -> None:
    """
    Associate relevant timing parameters to the innermost
    :py:func:`metacells.utilities.timing.timed_step`.

    The specified arguments are appended at the end of the generated ``timing.csv`` line. For
    example, ``timed_parameters(foo=2, bar=3)`` would add ``foo,2,bar,3`` to the line in
    ``timing.csv``.

    This allows tracking parameters which affect invocation time (such as array sizes), to help
    identify the causes for the long-running operations.
    """
    step_timing = current_step()
    if step_timing is not None:
        for name, value in kwargs.items():
            step_timing.parameters.append(name)
            step_timing.parameters.append(str(value))


CALLABLE = TypeVar("CALLABLE")


def timed_call(name: Optional[str] = None) -> Callable[[CALLABLE], CALLABLE]:
    """
    Automatically wrap each invocation of the decorated function with
    :py:func:`metacells.utilities.timing.timed_step` using the ``name`` (by default, the function's
    ``__qualname__``).

    Expected usage is:

    .. code:: python

        @ut.timed_call()
        def some_function(...):
            ...
    """
    if COLLECT_TIMING:

        def wrap(function: Callable) -> Callable:
            @wraps(function)
            def timed(*args: Any, **kwargs: Any) -> Any:
                with timed_step(name or function.__qualname__):
                    return function(*args, **kwargs)

            timed.__is_timed__ = True  # type: ignore
            return timed

    else:

        def wrap(function: Callable) -> Callable:
            function.__is_timed__ = True  # type: ignore
            return function

    return wrap  # type: ignore


def context() -> str:
    """
    Return the full current context (path of :py:func:`metacells.utilities.timing.timed_step`-s
    leading to the current point).

    .. note::

        The context will be the empty string unless we are actually collecting timing.
    """
    steps_stack = getattr(THREAD_LOCAL, "steps_stack", None)
    if not steps_stack:
        return ""
    return steps_stack[-1].context


def current_step() -> Optional[StepTiming]:
    """
    The timing collector for the innermost (current)
    :py:func:`metacells.utilities.timing.timed_step`, if any.
    """
    if not COLLECT_TIMING:
        return None
    steps_stack = getattr(THREAD_LOCAL, "steps_stack", None)
    if steps_stack is None or len(steps_stack) == 0:
        return None
    return steps_stack[-1]


# This is a circular dependency so having it at the end allows the exported symbols to be seen.



"""
Annotation
----------

In general we are using ``AnnData`` to hold the data being analyzed. However, the interface
of AnnData leaves some things out which are crucial for the proper working of our algorithm
(and any other algorithm that works at a scale of millions of cells).

X as an Annotation
..................

For a uniform interface, we pretend the ``X`` member is a per-variable-per-observation annotation
with the special name ``__x__``. This allows us to have APIs that take an annotation name and just
pass them (typically by default) the annotation "name" ``__x__`` to force the code to run on the
``X`` data member.

In general the APIs allow specifying either annotation names or alternatively an explicit matrix (or
vector for per-observation or per-variable annotations), for maximal usage flexibility.

Data Types
..........

The generic ``AnnData`` is cheerfully permissive when it comes to the data it contains. That is, when
accessing data, it isn't clear whether you'll be getting a numpy array or a pandas data series, and
for 2D data you might be getting all sort of data types (including sparse matrices of various
formats).

Python itself is very loose about the interface these data types provide - some operations such as
``len`` and ``shape`` and accessing an element by integer indices are safe, more advanced operations
can silently produce the wrong results, and most operations work on a subset of the data types,
often with wildly incompatible interfaces.

To combat this, we have the :py:mod:`metacells.utilities.typing` module which imposes some order on
the types zoo, and, in addition, we provide here accessor functions which return deterministic
usable data types, allowing for safe processing of the results. This is combined with the
py:mod:`metacells.utilities.computation` module which provides a set of operations that work
consistently on the few data types we use.

Data Layout
...........

A related issue is the layout of 2D data. For small matrices, this doesn't matter, but when dealing
with large matrices (millions of rows/columns), performing a simple operation may takes orders of
magnitude longer if applied to a matrix of the wrong layout.

To make things worse the builtin functions for converting between matrix layouts are pretty
inefficient so more efficient variants are provided in the py:mod:`metacells.utilities.computation`
module.

The accessors in this module allow for explicitly controlling the layout of the data they return,
and cache the different layouts of the same annotations of the ``AnnData`` (under the reasonable
assumption that the original data is not modified). This allows for writing
guaranteed-to-be-efficient processing code.

Data Logging
............

A side benefit of exclusively using the accessors provided here is that they participate in the
automated logging provided by the :py:mod:`metacells.utilities.logging` module. That is, using them
will automatically log writing the final results of a computation to the user at the ``INFO`` log
level, while higher logging level give insight into the exact data being read and written by the
algorithm's nested sub-steps.
"""

from typing import Any
from typing import Callable
from typing import Collection
from typing import Dict
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import Tuple
from typing import Union

import anndata as ad  # type: ignore
import numpy as np
from anndata import AnnData

import metacells.utilities.computation as utc
import metacells.utilities.logging as utl
import metacells.utilities.timing as utm
import metacells.utilities.typing as utt

__all__ = [
    "slice",
    "copy_adata",
    "set_name",
    "get_name",
    "set_m_data",
    "get_m_data",
    "set_o_data",
    "get_o_series",
    "get_o_numpy",
    "get_o_names",
    "maybe_o_numpy",
    "set_v_data",
    "get_v_series",
    "get_v_numpy",
    "get_v_names",
    "maybe_v_numpy",
    "set_oo_data",
    "get_oo_frame",
    "get_oo_proper",
    "set_vv_data",
    "get_vv_frame",
    "get_vv_proper",
    "set_oa_data",
    "get_oa_frame",
    "get_oa_proper",
    "set_va_data",
    "get_va_frame",
    "get_va_proper",
    "set_vo_data",
    "get_vo_frame",
    "get_vo_proper",
    "has_data",
]


Annotations = Union[MutableMapping[Any, Any], utt.PandasFrame]


@utm.timed_call()
def slice(  # pylint: disable=redefined-builtin,too-many-branches,too-many-statements
    adata: AnnData,
    *,
    name: Optional[str] = None,
    obs: Optional[utt.Vector] = None,
    vars: Optional[utt.Vector] = None,
    track_obs: Optional[str] = None,
    track_var: Optional[str] = None,
    share_derived: bool = True,
    top_level: bool = True,
) -> AnnData:
    """
    Return new annotated data which includes a subset of the full ``adata``.

    If ``name`` is not specified, the data will be unnamed. Otherwise, if it starts with a ``.``, it
    will be appended to the current name (if any). Otherwise, ``name`` is the new name.

    If ``obs`` and/or ``vars`` are specified, they should be set to either a boolean mask or a
    collection of indices to include in the data slice. In the case of an indices array, it is
    assumed the indices are unique and sorted, that is that their effect is similar to a mask.

    If ``track_obs`` and/or ``track_var`` are specified, the result slice will include a
    per-observation and/or per-variable annotation containing the indices of the sliced elements in
    the original full data.

    If the slice happens to be the full original data, then this becomes equivalent to
    :py:func:`copy_adata`, and by default this will ``share_derived`` (share the derived data
    cache).
    """
    assert "__x__" not in adata.layers

    is_same_obs: Optional[bool] = None
    if obs is None:
        obs = range(adata.n_obs)
        is_same_obs = True
    else:
        assert 0 < len(obs) <= adata.n_obs
        if len(obs) < adata.n_obs:
            is_same_obs = False

        obs = utt.to_numpy_vector(obs)
        if obs.dtype == "bool":
            assert obs.size == adata.n_obs
            assert np.any(obs)
            is_same_obs = bool(np.all(obs))

    is_same_vars: Optional[bool] = None
    if vars is None:
        vars = range(adata.n_vars)
        is_same_vars = True
    else:
        assert 0 < len(vars) <= adata.n_vars
        if len(vars) < adata.n_vars:
            is_same_vars = False

        vars = utt.to_numpy_vector(vars)
        if vars.dtype == "bool":
            assert vars.size == adata.n_vars
            assert np.any(vars)
            is_same_vars = bool(np.all(vars))

    if is_same_obs and is_same_vars:
        bdata = copy_adata(adata, name=name, share_derived=share_derived, top_level=top_level)

    else:
        if is_same_vars:
            replaced = _replace_with_layout(adata, "row_major")
        elif is_same_obs:
            replaced = _replace_with_layout(adata, "column_major")
        else:
            replaced = {}

        try:
            with utm.timed_step("adata.slice"):
                bdata = adata[obs, vars].copy()
        finally:
            _replace_back(adata, replaced)

        set_name(bdata, name)

        if hasattr(bdata, "__is_top_level__"):
            delattr(bdata, "__is_top_level__")
        if hasattr(bdata, "__derived__"):
            delattr(bdata, "__derived__")

        if share_derived and (is_same_obs or is_same_obs is None) and (is_same_vars or is_same_vars is None):
            if is_same_obs is None:
                is_same_obs = bdata.n_obs == adata.n_obs and bool(np.all(bdata.obs_names == adata.obs_names))

            if is_same_vars is None:
                is_same_vars = bdata.n_vars == adata.n_vars and bool(np.all(bdata.var_names == adata.var_names))

            if is_same_obs and is_same_vars:
                if not hasattr(adata, "__derived__"):
                    setattr(adata, "__derived__", {})
                setattr(bdata, "__derived__", getattr(adata, "__derived__"))

        if top_level:
            utl.top_level(bdata)

    if utl.logging_calc():
        utl.log_calc(f'slice {get_name(adata, "unnamed")} into {get_name(bdata, "unnamed")} shape {bdata.shape}')  #

    if track_obs is not None:
        set_o_data(bdata, track_obs, np.arange(adata.n_obs, dtype="int32")[obs])

    if track_var is not None:
        set_v_data(bdata, track_var, np.arange(adata.n_vars, dtype="int32")[vars])

    return bdata


@utm.timed_call()
def _replace_with_layout(adata: AnnData, layout: str) -> Dict[str, utt.Matrix]:
    replaced: Dict[str, utt.Matrix] = {}

    matrix: utt.Matrix = adata.X
    if not utt.is_layout(matrix, layout):
        replaced["__x__"] = matrix
        adata.X = get_vo_proper(adata, "__x__", layout=layout)

    for name in adata.layers:
        matrix = adata.layers[name]
        if not utt.is_layout(matrix, layout):
            replaced[name] = matrix
            adata.layers[name] = get_vo_proper(adata, name, layout=layout)

    return replaced


@utm.timed_call()
def _replace_back(adata: AnnData, replaced: Dict[str, utt.Matrix]) -> None:
    for name, matrix in replaced.items():
        if name == "__x__":
            adata.X = matrix
        else:
            adata.layers[name] = matrix


@utm.timed_call()
def copy_adata(
    adata: AnnData, *, name: Optional[str] = None, share_derived: bool = True, top_level: bool = True
) -> AnnData:
    """
    Return a copy of some annotated ``adata``.

    If ``name`` is not specified, the data will be unnamed. Otherwise, if it starts with a ``.``, it
    will be appended to the current name (if any). Otherwise, ``name`` is the new name.

    If ``share_derived`` is ``True`` (the default), then the copy will share the derived data cache,
    which contains specific layout variants of matrix data and sums of columns/rows of matrix data.
    Use this if you intend to modify the copy in-place.

    .. note::

        In general we assume annotated data is **not** modified in-place, but it might make sense to
        create a copy (**not** sharing derived data), modify it immediately (before accessing data
        in a specific layout), and then proceed to process it without further modifications.
    """
    with utm.timed_step("adata.copy"):
        bdata = adata.copy()

    set_name(bdata, name)

    if hasattr(bdata, "__is_top_level__"):
        delattr(bdata, "__is_top_level__")

    if share_derived:
        if not hasattr(adata, "__derived__"):
            setattr(adata, "__derived__", {})
        setattr(bdata, "__derived__", getattr(adata, "__derived__"))
    else:
        if hasattr(bdata, "__derived__"):
            delattr(bdata, "__derived__")

    if top_level:
        utl.top_level(bdata)
    return bdata


def has_data(
    adata: AnnData,
    name: str,
    layout: Optional[str] = None,
) -> bool:
    """
    Test whether we have the specified data.

    If the data is per-variable-per-observation, and ``layout`` is specified (one of ``row_major``
    and ``column_major``), it returns whether the specific data layout is available in the cache,
    without having to re-layout existing data.
    """
    assert layout is None or layout in utt.LAYOUT_OF_AXIS

    derived: Union[Dict[str, Any], List[str]] = []
    if hasattr(adata, "__derived__"):
        derived = getattr(adata, "__derived__")

    if name == "__x__":
        if layout is None:
            return True
        matrix: utt.Matrix = adata.X
        return utt.is_layout(matrix, layout) or f"vo:__x__:{layout}" in derived

    for per, annotations in (
        ("vo", adata.layers),
        ("oo", adata.obsp),
        ("vv", adata.varp),
        ("va", adata.obsm),
        ("oa", adata.varm),
    ):
        if name not in annotations:
            continue
        if layout is None:
            return True
        matrix = annotations[name]
        return utt.is_layout(matrix, layout) or f"{per}:{name}:{layout}" in derived

    assert layout is None

    for annotations in (adata.obs, adata.var, adata.uns):
        if name in annotations:
            return True

    return False


def get_m_data(
    adata: AnnData,
    name: str,
    *,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    """
    Get metadata (unstructured annotation) in ``adata`` by its ``name``.
    """
    data = adata.uns.get(name)
    if data is None:
        raise _unknown_data(adata, name, "m")
    utl.log_get(adata, "m", name, data, formatter=formatter)
    return data


def _get_o_data(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    if not isinstance(name, str):
        log_name = "<data>"
    else:
        if name.endswith("|sum"):
            name = name[:-4]
            assert not sum
            sum = True
        log_name = name

    if sum:
        if formatter is None:
            formatter = utl.sizes_description
        data = _get_vo_sum_data(adata, "row", name)
        log_name += "|sum"
    else:
        data = _get_shaped_data(adata, "o", adata.obs, shape=(adata.n_obs,), name=name)
    utl.log_get(adata, "o", log_name, data, formatter=formatter)
    return data


def get_o_series(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasSeries:
    """
    Get per-observation (cell) data in ``adata`` by its ``name`` as a pandas series.

    If ``name`` is a string, it is the name of a per-observation annotation to fetch.
    Otherwise, it should be some vector of data of the appropriate size.
    """
    data = _get_o_data(adata, name, sum=sum, formatter=formatter)
    series = utt.maybe_pandas_series(data)
    if series is None:
        series = utt.to_pandas_series(data, index=adata.obs_names)
    return series


def get_o_numpy(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.NumpyVector:
    """
    Get per-observation (cell) data in ``adata`` by its ``name`` as a Numpy array.

    If ``name`` is a string, it is the name of a per-observation annotation to fetch.
    Otherwise, it should be some vector of data of the appropriate size.

    If ``sum`` is ``True``, then ``name`` should be the name of a per-observation-per-variable
    annotation, or a matrix, and this will return the sum (per row) of this data.
    """
    data = _get_o_data(adata, name, sum=sum, formatter=formatter)
    return utt.to_numpy_vector(data)


def get_o_names(
    adata: AnnData,
) -> utt.NumpyVector:
    """
    Get a numpy vector of observation names.
    """
    data = utt.to_numpy_vector(adata.obs_names)
    utl.log_get(adata, "o", "__name__", data)
    return data


def maybe_o_numpy(
    adata: AnnData,
    name: Union[str, utt.Shaped, None],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Optional[utt.NumpyVector]:
    """
    Similar to :py:func:`get_o_numpy`, but if ``name`` is ``None``, return ``None``.
    """
    if name is None:
        return None
    return get_o_numpy(adata, name, sum=sum, formatter=formatter)


def _get_v_data(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    if not isinstance(name, str):
        log_name = "<data>"
    else:
        if isinstance(name, str) and name.endswith("|sum"):
            name = name[:-4]
            assert not sum
            sum = True
        log_name = name

    if sum:
        if formatter is None:
            formatter = utl.sizes_description
        data = _get_vo_sum_data(adata, "column", name)
        log_name += "|sum"
    else:
        data = _get_shaped_data(adata, "v", adata.var, shape=(adata.n_vars,), name=name)
    utl.log_get(adata, "v", log_name, data, formatter=formatter)
    return data


def get_v_series(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasSeries:
    """
    Get per-variable (gene) data in ``adata`` by its ``name`` as a pandas series.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some vector of data of the appropriate size.
    """
    data = _get_v_data(adata, name, sum=sum, formatter=formatter)
    series = utt.maybe_pandas_series(data)
    if series is None:
        series = utt.to_pandas_series(data, index=adata.var_names)
    return series


def get_v_numpy(
    adata: AnnData,
    name: Union[str, utt.Shaped],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.NumpyVector:
    """
    Get per-variable (gene) data in ``adata`` by its ``name`` as a numpy array.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some vector of data of the appropriate size.
    """
    data = _get_v_data(adata, name, sum=sum, formatter=formatter)
    return utt.to_numpy_vector(data)


def get_v_names(
    adata: AnnData,
) -> utt.NumpyVector:
    """
    Get a numpy vector of variable names.
    """
    data = utt.to_numpy_vector(adata.var_names)
    utl.log_get(adata, "v", "__name__", data)
    return data


def maybe_v_numpy(
    adata: AnnData,
    name: Union[str, utt.Shaped, None],
    *,
    sum: bool = False,  # pylint: disable=redefined-builtin
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Optional[utt.NumpyVector]:
    """
    Similar to :py:func:`get_v_numpy`, but if ``name`` is ``None``, return ``None``.
    """
    if name is None:
        return None
    return get_v_numpy(adata, name, sum=sum, formatter=formatter)


def _get_oo_data(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.Matrix:
    data = _get_layout_data(adata, "oo", adata.obsp, shape=(adata.n_obs, adata.n_obs), name=name, layout=layout)
    utl.log_get(adata, "oo", name, data, formatter=formatter)
    return data


def get_oo_frame(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasFrame:
    """
    Get per-observation-per-observation (per-cell-per-cell) data as a pandas data frame.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some matrix of data of the appropriate size.

    If ``layout`` (default: {layout}) is specified, it must be one of ``row_major`` or
    ``column_major``. If this requires relayout of the data, the result is cached in a hidden data
    member for future reuse.

    .. note::

        Since Pandas frames do not play well with sparse representations, this will **always** return the
        data as a dense matrix. For very large data, this may consume much more memory, so use with care.
    """
    data = _get_oo_data(adata, name, layout=layout, formatter=formatter)
    frame = utt.maybe_pandas_frame(data)
    if frame is None:
        frame = utt.to_pandas_frame(utt.to_numpy_matrix(data), index=adata.obs_names, columns=adata.obs_names)
    return frame


def get_oo_proper(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.ProperMatrix:
    """
    Same as ``get_oo_data`` but returns a :py:const:`metacells.utilities.typing.ProperMatrix`.
    """
    data = _get_oo_data(adata, name, layout=layout, formatter=formatter)
    return utt.to_proper_matrix(data, default_layout=layout or "row_major")


def _get_vv_data(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.Matrix:
    data = _get_layout_data(adata, "vv", adata.varp, shape=(adata.n_vars, adata.n_vars), name=name, layout=layout)
    utl.log_get(adata, "vv", name, data, formatter=formatter)
    return data


def get_vv_frame(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasFrame:
    """
    Get per-variable-per-variable (per-gene-per-gene) data as a pandas data frame.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some matrix of data of the appropriate size.

    If ``layout`` (default: {layout}) is specified, it must be one of ``row_major`` or
    ``column_major``. If this requires relayout of the data, the result is cached in a hidden data
    member for future reuse.

    .. note::

        Since Pandas frames do not play well with sparse representations, this will **always** return the
        data as a dense matrix. For very large data, this may consume much more memory, so use with care.
    """
    data = _get_vv_data(adata, name, layout=layout, formatter=formatter)
    frame = utt.maybe_pandas_frame(data)
    if frame is None:
        frame = utt.to_pandas_frame(utt.to_numpy_matrix(data), index=adata.var_names, columns=adata.var_names)
    return frame


def get_vv_proper(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.ProperMatrix:
    """
    Same as ``get_vv_data`` but returns a :py:const:`metacells.utilities.typing.ProperMatrix`.
    """
    data = _get_vv_data(adata, name, layout=layout, formatter=formatter)
    return utt.to_proper_matrix(data, default_layout=layout or "row_major")


def _get_oa_data(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.Matrix:
    data = _get_layout_data(adata, "oa", adata.obsm, shape=(adata.n_obs, 0), name=name, layout=layout)
    utl.log_get(adata, "oa", name, data, formatter=formatter)
    return data


def get_oa_frame(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    columns: Optional[Collection],
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasFrame:
    """
    Get per-observation-per-any (per-cell-per-any) data as a pandas data frame.

    Rows are observations (cells), indexed by the observation names (typically cell barcode).
    Columns are "something" - specify ``columns`` to specify an index.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some matrix of data of the appropriate size.

    If ``layout`` (default: {layout}) is specified, it must be one of ``row_major`` or
    ``column_major``. If this requires relayout of the data, the result is cached in a hidden data
    member for future reuse.

    .. note::

        Since Pandas frames do not play well with sparse representations, this will **always** return the
        data as a dense matrix. For very large data, this may consume much more memory, so use with care.
    """
    data = _get_oa_data(adata, name, layout=layout, formatter=formatter)
    frame = utt.maybe_pandas_frame(data)
    if frame is None:
        frame = utt.to_pandas_frame(utt.to_numpy_matrix(data), index=adata.obs_names, columns=columns)
    return frame


def get_oa_proper(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.ProperMatrix:
    """
    Same as ``get_oa_data`` but returns a :py:const:`metacells.utilities.typing.ProperMatrix`.
    """
    data = _get_oa_data(adata, name, layout=layout, formatter=formatter)
    return utt.to_proper_matrix(data, default_layout=layout or "row_major")


def _get_va_data(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.Matrix:
    data = _get_layout_data(adata, "va", adata.varm, shape=(adata.n_vars, 0), name=name, layout=layout)
    utl.log_get(adata, "va", name, data, formatter=formatter)
    return data


def get_va_frame(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    columns: Optional[Collection],
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasFrame:
    """
    Get per-variable-per-any (per-cell-per-any) data as a pandas data frame.

    Rows are variables (genes), indexed by their names.
    Columns are "something" - specify ``columns`` to specify an index.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some matrix of data of the appropriate size.

    If ``layout`` (default: {layout}) is specified, it must be one of ``row_major`` or
    ``column_major``. If this requires relayout of the data, the result is cached in a hidden data
    member for future reuse.

    .. note::

        Since Pandas frames do not play well with sparse representations, this will **always** return the
        data as a dense matrix. For very large data, this may consume much more memory, so use with care.
    """
    data = _get_va_data(adata, name, layout=layout, formatter=formatter)
    frame = utt.maybe_pandas_frame(data)
    if frame is None:
        frame = utt.to_pandas_frame(utt.to_numpy_matrix(data), index=adata.var_names, columns=columns)
    return frame


def get_va_proper(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.ProperMatrix:
    """
    Same as ``get_va_data`` but returns a :py:const:`metacells.utilities.typing.ProperMatrix`.
    """
    data = _get_va_data(adata, name, layout=layout, formatter=formatter)
    return utt.to_proper_matrix(data, default_layout=layout or "row_major")


def _get_vo_data(
    adata: AnnData,
    name: Union[str, utt.Matrix],
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.Matrix:
    if name is None:
        name = "__x__"

    data = _get_layout_data(adata, "vo", adata.layers, shape=(adata.n_obs, adata.n_vars), name=name, layout=layout)
    utl.log_get(adata, "vo", name, data, formatter=formatter)
    return data


def get_vo_frame(
    adata: AnnData,
    name: Union[str, utt.Matrix] = "__x__",
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.PandasFrame:
    """
    Get per-variable-per-observation (per-gene-per-cell) data as a pandas data frame.

    Rows are observations (cells), indexed by the observation names (typically cell barcode).
    Columns are variables (genes), indexed by their names.

    If ``name`` is a string, it is the name of a per-variable annotation to fetch.
    Otherwise, it should be some matrix of data of the appropriate size.

    If ``layout`` (default: {layout}) is specified, it must be one of ``row_major`` or
    ``column_major``. If this requires relayout of the data, the result is cached in a hidden data
    member for future reuse.

    .. note::

        Since Pandas frames do not play well with sparse representations, this will **always** return the
        data as a dense matrix. For very large data, this may consume much more memory, so use with care.
    """
    data = _get_vo_data(adata, name, layout=layout, formatter=formatter)
    frame = utt.maybe_pandas_frame(data)
    if frame is None:
        frame = utt.to_pandas_frame(utt.to_numpy_matrix(data), index=adata.obs_names, columns=adata.var_names)
    return frame


def get_vo_proper(
    adata: AnnData,
    name: Union[str, utt.Matrix] = "__x__",
    *,
    layout: Optional[str] = None,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> utt.ProperMatrix:
    """
    Same as ``get_vo_data`` but returns a :py:const:`metacells.utilities.typing.ProperMatrix`.
    """
    data = _get_vo_data(adata, name, layout=layout, formatter=formatter)
    return utt.to_proper_matrix(data, default_layout=layout or "row_major")


def _get_vo_sum_data(
    adata: AnnData,
    per: str,
    name: Union[str, utt.Shaped],
) -> utt.NumpyVector:
    if not isinstance(name, str):
        assert utt.is_2d(name)
        assert name.shape == adata.shape  # type: ignore
        return utc.sum_per(name, per=per)  # type: ignore

    if not hasattr(adata, "__derived__"):
        derived: Dict[str, utt.NumpyVector] = {}
        setattr(adata, "__derived__", derived)
    else:
        derived = getattr(adata, "__derived__")

    sum_name = f"vo:{name}:sum_per_{per}"
    sum_data = derived.get(sum_name)
    if sum_data is not None:
        return sum_data

    data = get_vo_proper(adata, name, layout=f"{per}_major")
    sum_data = utc.sum_per(data, per=per)
    derived[sum_name] = sum_data
    return sum_data


def _get_layout_data(
    adata: AnnData,
    per: str,
    annotations: Annotations,
    *,
    shape: Tuple[int, ...],
    name: Union[str, utt.Matrix],
    layout: Optional[str],
) -> Any:
    data = _get_shaped_data(adata, per, annotations, shape=shape, name=name)
    if not isinstance(name, str):
        if not utt.is_layout(data, layout):
            assert layout is not None
            data = utc.to_layout(name, layout=layout)
        return data

    if utt.is_layout(data, layout):
        return data

    assert layout in utt.LAYOUT_OF_AXIS
    layout_name = f"{per}:{name}:{layout}"

    if not hasattr(adata, "__derived__"):
        derived: Dict[str, utt.ProperMatrix] = {}
        setattr(adata, "__derived__", derived)
    else:
        derived = getattr(adata, "__derived__")

    layout_data = derived.get(layout_name)
    if layout_data is not None:
        assert layout_data.shape == shape
        assert utt.is_layout(layout_data, layout)
    else:
        layout_data = utc.to_layout(data, layout=layout)
        derived[layout_name] = layout_data

    if not utt.frozen(layout_data):
        utt.freeze(layout_data)

    return layout_data


def _get_shaped_data(
    adata: AnnData,
    per: str,
    annotations: Annotations,
    *,
    shape: Tuple[int, ...],
    name: Union[str, utt.Shaped],
) -> Any:
    if isinstance(name, str):
        if per == "vo" and name == "__x__":
            data = _fix_data(adata.X)
        else:
            if name not in annotations:  # type: ignore
                raise _unknown_data(adata, name, per)
            data = _fix_data(annotations[name])

        if not utt.frozen(data):
            utt.freeze(data)

    else:
        if utt.is_1d(name):
            data = utt.to_numpy_vector(name)
        else:
            data = utt.to_proper_matrix(name)  # type: ignore

    assert data.shape == shape
    return data


def _fix_data(data: Any) -> Any:
    if isinstance(data, ad._core.sparse_dataset.SparseDataset):  # pylint: disable=protected-access
        return data.value
    return data


def _annotation_per(adata: AnnData, per: str) -> Annotations:  # pylint: disable=too-many-return-statements
    if per == "m":
        return adata.uns
    if per == "vo":
        return adata.layers
    if per == "v":
        return adata.var
    if per == "o":
        return adata.obs
    if per == "vv":
        return adata.varp
    if per == "oo":
        return adata.obsp
    if per == "va":
        return adata.varm
    if per == "oa":
        return adata.obsm
    raise ValueError(f"unknown per: {per}")


def set_name(adata: AnnData, name: Optional[str]) -> None:
    """
    Set the ``name`` of the data (for log messages).

    If the name starts with ``.`` it is appended to the current name, if any.
    """
    if name is None:
        if "__name__" in adata.uns:
            del adata.uns["__name__"]
        return

    if name[0] == ".":
        old_name = get_name(adata)
        if old_name is None:
            name = name[1:]
        else:
            name = old_name + name
    adata.uns["__name__"] = name


def get_name(adata: AnnData, default: Optional[str] = None) -> Optional[str]:
    """
    Return the name of the data (for log messages), if any.

    If no name was set, returns the ``default``.
    """
    return adata.uns.get("__name__", default)


def set_m_data(
    adata: AnnData,
    name: str,
    data: Any,
    *,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    """
    Set unstructured data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "m", name, data, formatter=formatter)
    adata.uns[name] = data


def set_o_data(
    adata: AnnData,
    name: str,
    data: utt.NumpyVector,
    *,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    """
    Set per-observation (cell) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "o", name, data, formatter=formatter)

    if not isinstance(data, list):
        utt.mustbe_canonical(data)
        if not utt.frozen(data):
            utt.freeze(data)

    adata.obs[name] = data


def set_v_data(
    adata: AnnData, name: str, data: utt.NumpyVector, *, formatter: Optional[Callable[[Any], Any]] = None
) -> Any:
    """
    Set per-variable (gene) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "v", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    adata.var[name] = data


def set_oo_data(
    adata: AnnData, name: str, data: utt.ProperMatrix, *, formatter: Optional[Callable[[Any], Any]] = None
) -> Any:
    """
    Set per-observation-per-observation (cell) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "oo", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    adata.obsp[name] = data


def set_vv_data(
    adata: AnnData, name: str, data: utt.ProperMatrix, *, formatter: Optional[Callable[[Any], Any]] = None
) -> Any:
    """
    Set per-variable-per-variable (gene) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "vv", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    adata.varp[name] = data


def set_oa_data(
    adata: AnnData, name: str, data: utt.ProperMatrix, *, formatter: Optional[Callable[[Any], Any]] = None
) -> Any:
    """
    Set per-observation-per-any (cell) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "oa", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    adata.obsm[name] = data


def set_va_data(
    adata: AnnData, name: str, data: utt.ProperMatrix, *, formatter: Optional[Callable[[Any], Any]] = None
) -> Any:
    """
    Set per-variable-per-any (gene) data.

    If ``formatter`` is specified, its results is used when logging the operation.
    """
    utl.log_set(adata, "va", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    adata.varm[name] = data


def set_vo_data(
    adata: AnnData,
    name: str,
    data: utt.ProperMatrix,
    *,
    formatter: Optional[Callable[[Any], Any]] = None,
) -> Any:
    """
    Set per-variable-per-observation (per-gene-per-cell) data.
    """
    utl.log_set(adata, "vo", name, data, formatter=formatter)

    utt.mustbe_canonical(data)
    if not utt.frozen(data):
        utt.freeze(data)

    if name == "__x__":
        adata.X = data
    else:
        adata.layers[name] = data

    if hasattr(adata, "__derived__"):
        derived = getattr(adata, "__derived__")
        marker_name = f":{name}:"
        deleted_names: List[str] = []
        for derived_name in derived.keys():
            if marker_name in derived_name:
                deleted_names.append(derived_name)
        for deleted_name in deleted_names:
            del derived[deleted_name]


def _unknown_data(adata: AnnData, name: str, per: Optional[str] = None) -> KeyError:
    texts = ["unknown"]

    if per is not None:
        texts.append(" ")
        texts.append(per)

    texts.append(" data")

    data_name = get_name(adata)
    if data_name is not None:
        texts.append(": ")
        texts.append(data_name)

    texts.append(" name: ")
    texts.append(name)
    return KeyError("".join(texts))



"""
Test the utility functions.
"""

import os
from glob import glob
from typing import Any
from typing import List

import numpy as np
from scipy import sparse  # type: ignore
from scipy import stats
from sklearn.metrics import roc_auc_score  # type: ignore

import metacells.utilities as ut

ut.allow_inefficient_layout(False)

# pylint: disable=missing-function-docstring


def test_expand_doc() -> None:
    @ut.expand_doc(foo=7)
    def bar(baz: Any, vaz: int = 5) -> None:  # pylint: disable=blacklisted-name,unused-argument
        """
        Bar with {foo} foos and parameter vaz (default: {vaz}).
        """

    assert (
        bar.__doc__
        == """
        Bar with 7 foos and parameter vaz (default: 5).
        """
    )


def test_corrcoef() -> None:
    np.random.seed(123456)
    matrix = sparse.rand(101, 10101, density=0.1, format="csr")
    # matrix = sparse.rand(5, 20, density=0.3, format='csr')
    dense = ut.to_layout(matrix.toarray(), layout="row_major")
    numpy_correlation = np.corrcoef(dense)
    assert numpy_correlation.shape == (101, 101)

    dense = dense.T

    for reproducible in (False, True):
        dense_correlation = ut.corrcoef(dense, per="column", reproducible=reproducible)
        assert dense_correlation.shape == (101, 101)
        assert np.min(np.diag(dense_correlation)) == 1
        assert np.max(np.diag(dense_correlation)) == 1
        assert np.allclose(dense_correlation, numpy_correlation, atol=1e-6)

        sparse_correlation = ut.corrcoef(matrix, per="row", reproducible=reproducible)
        assert sparse_correlation.shape == (101, 101)
        assert np.min(np.diag(sparse_correlation)) == 1
        assert np.max(np.diag(sparse_correlation)) == 1
        assert np.allclose(sparse_correlation, numpy_correlation, atol=1e-6)

    dense[:, :] = 0
    for reproducible in (False, True):
        zeros_correlation = ut.corrcoef(dense, per="column", reproducible=reproducible)
        assert zeros_correlation.shape == (101, 101)
        assert np.min(np.diag(zeros_correlation)) == 1
        assert np.max(np.diag(zeros_correlation)) == 1
        np.fill_diagonal(zeros_correlation, 0)
        assert np.min(zeros_correlation) == 0
        assert np.max(zeros_correlation) == 0


def test_cross_corrcoef() -> None:
    np.random.seed(123456)
    first_matrix = ut.to_numpy_matrix(sparse.rand(51, 10101, density=0.1, format="csr"))
    second_matrix = ut.to_numpy_matrix(sparse.rand(101, 10101, density=0.1, format="csr"))

    first_dense = ut.to_layout(first_matrix, layout="row_major")
    second_dense = ut.to_layout(second_matrix, layout="row_major")
    fast_results = ut.cross_corrcoef_rows(first_dense, second_dense, reproducible=True)

    slow_results = np.zeros((51, 101), dtype="float")
    for row in range(51):
        for column in range(101):
            slow_results[row, column] = np.corrcoef(first_dense[row, :], second_dense[column, :])[0, 1]
    slow_results[np.isnan(slow_results)] = 0

    assert fast_results.shape == slow_results.shape
    assert np.allclose(fast_results, slow_results, atol=1e-6)


def test_pairs_corrcoef() -> None:
    np.random.seed(123456)
    first_matrix = ut.to_numpy_matrix(sparse.rand(100, 100, density=0.1, format="csr"))
    second_matrix = ut.to_numpy_matrix(sparse.rand(100, 100, density=0.1, format="csr"))

    first_dense = ut.to_layout(first_matrix, layout="row_major")
    second_dense = ut.to_layout(second_matrix, layout="row_major")
    fast_results = ut.pairs_corrcoef_rows(first_dense, second_dense, reproducible=True)

    slow_results = np.zeros(100, dtype="float")
    for row in range(100):
        slow_results[row] = np.corrcoef(first_dense[row], second_dense[row])[0, 1]
    slow_results[np.isnan(slow_results)] = 0

    assert fast_results.shape == slow_results.shape
    assert np.allclose(fast_results, slow_results, atol=1e-6)


def test_logistics() -> None:
    matrix = np.array([[0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1], [0, 0, 0, 1, 1, 1]], dtype="float64")
    same_value = 1 / (1 + np.exp(-5 * (0 - 0.8)))
    scale = 1 / (1 - same_value)
    diff_value = 1 / (1 + np.exp(-5 * (1 - 0.8)))
    normalized_value = (diff_value - same_value) * scale
    results = ut.logistics(matrix, per="row", location=0.8, slope=5)
    assert np.allclose(results[0, 0], 0)
    assert np.allclose(results[1, 1], 0)
    assert np.allclose(results[2, 2], 0)
    assert np.allclose(results[0, 1], normalized_value)
    assert np.allclose(results[1, 0], normalized_value)
    assert np.allclose(results[0, 2], normalized_value / 2)
    assert np.allclose(results[2, 0], normalized_value / 2)
    assert np.allclose(results[1, 2], normalized_value / 2)
    assert np.allclose(results[2, 1], normalized_value / 2)


def test_cross_logistics() -> None:
    first_matrix = np.array([[0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1]], dtype="float64")
    second_matrix = np.array([[1, 1, 1, 1, 1, 1], [0, 0, 0, 1, 1, 1]], dtype="float64")
    same_value = 1 / (1 + np.exp(-5 * (0 - 0.8)))
    scale = 1 / (1 - same_value)
    diff_value = 1 / (1 + np.exp(-5 * (1 - 0.8)))
    normalized_value = (diff_value - same_value) * scale
    results = ut.cross_logistics_rows(first_matrix, second_matrix, location=0.8, slope=5)

    assert np.allclose(results[0, 0], normalized_value)
    assert np.allclose(results[0, 1], normalized_value / 2)
    assert np.allclose(results[1, 1], normalized_value / 2)
    assert np.allclose(results[1, 0], 0)


def test_pairs_logistics() -> None:
    first_matrix = np.array([[0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1]], dtype="float64")
    second_matrix = np.array([[1, 1, 1, 1, 1, 1], [0, 0, 0, 1, 1, 1]], dtype="float64")
    same_value = 1 / (1 + np.exp(-5 * (0 - 0.8)))
    scale = 1 / (1 - same_value)
    diff_value = 1 / (1 + np.exp(-5 * (1 - 0.8)))
    normalized_value = (diff_value - same_value) * scale
    results = ut.pairs_logistics_rows(first_matrix, second_matrix, location=0.8, slope=5)

    assert np.allclose(results[0], normalized_value)
    assert np.allclose(results[1], normalized_value / 2)


def test_relayout_matrix() -> None:
    rvs = stats.poisson(10, loc=10).rvs
    csr_matrix = sparse.random(20, 20, format="csr", dtype="int32", random_state=123456, data_rvs=rvs)
    assert csr_matrix.getformat() == "csr"

    scipy_csc_matrix = csr_matrix.tocsc()
    metacells_csc_matrix = ut.to_layout(csr_matrix, layout="column_major")

    assert scipy_csc_matrix.getformat() == "csc"
    assert metacells_csc_matrix.getformat() == "csc"

    assert np.all(metacells_csc_matrix.indptr == scipy_csc_matrix.indptr)
    assert np.all(metacells_csc_matrix.toarray() == scipy_csc_matrix.toarray())

    scipy_csr_matrix = scipy_csc_matrix.tocsr()

    metacells_csr_matrix = ut.to_layout(metacells_csc_matrix, layout="row_major")

    assert scipy_csr_matrix.getformat() == "csr"
    assert metacells_csr_matrix.getformat() == "csr"

    assert np.all(metacells_csr_matrix.indptr == scipy_csr_matrix.indptr)
    assert np.all(metacells_csr_matrix.toarray() == scipy_csr_matrix.toarray())


def test_downsample_vector_inplace() -> None:
    size = 10
    samples = 20
    random_seed = 123456

    safe = np.arange(size)
    data = np.arange(size)

    ut.downsample_vector(data, samples, random_seed=random_seed)

    assert np.all(data <= safe)
    assert np.sum(data) == samples


def test_downsample_vector_too_few() -> None:
    size = 10
    total = (size * (size - 1)) // 2
    samples = total * 2
    random_seed = 123456

    safe = np.arange(size)
    data = np.arange(size)

    ut.downsample_vector(data, samples, random_seed=random_seed)

    assert np.all(data == safe)


def test_downsample_matrix() -> None:
    rvs = stats.poisson(10, loc=10).rvs
    matrix = sparse.random(1000, 10000, format="csr", dtype="int32", random_state=123456, data_rvs=rvs)
    assert matrix.nnz == matrix.shape[0] * matrix.shape[1] * 0.01
    old_row_sums = ut.sum_per(matrix, per="row")
    min_sum = np.min(old_row_sums)
    result = ut.downsample_matrix(matrix, per="row", samples=int(min_sum), random_seed=123456)
    assert result.shape == matrix.shape
    new_row_sums = ut.sum_per(result, per="row")
    assert np.all(new_row_sums == min_sum)

    matrix = matrix.toarray()
    result = ut.downsample_matrix(matrix, per="row", samples=int(min_sum), random_seed=123456)
    assert result.shape == matrix.shape
    new_row_sums = ut.sum_per(result, per="row")
    assert np.all(new_row_sums == min_sum)


def test_matrix_rows_folds_and_aurocs() -> None:
    np.random.seed(123456)

    signal = np.random.rand(100)
    zero_mask = signal < 1 / 3
    one_mask = signal > 2 / 3
    signal[zero_mask] = 0

    dense = np.random.rand(20, 100)
    dense[:, zero_mask] = 0
    for index in range(20):
        dense[index, :] *= index / 10
        dense[index, :] += signal

    in_data = ut.to_layout(dense[:, one_mask], layout="row_major")
    out_data = ut.to_layout(dense[:, ~one_mask], layout="row_major")
    in_means = ut.mean_per(in_data, per="row")
    out_means = ut.mean_per(out_data, per="row")

    normalization = 1e-4
    in_means += normalization
    out_means += normalization
    builtin_rows_folds = in_means / out_means

    builtin_rows_aurocs = np.array([roc_auc_score(one_mask, dense[index, :]) for index in range(20)])

    dense_rows_folds, dense_rows_aurocs = ut.matrix_rows_folds_and_aurocs(
        dense, columns_subset=one_mask, normalization=normalization
    )
    assert np.allclose(dense_rows_aurocs, builtin_rows_aurocs)
    assert np.allclose(dense_rows_folds, builtin_rows_folds)

    compressed = sparse.csr_matrix(dense)
    sparse_rows_folds, sparse_rows_aurocs = ut.matrix_rows_folds_and_aurocs(
        compressed, columns_subset=one_mask, normalization=normalization
    )
    assert np.allclose(sparse_rows_aurocs, builtin_rows_aurocs)
    assert np.allclose(sparse_rows_folds, builtin_rows_folds)


def test_bincount_vector() -> None:
    array = np.array(np.random.rand(100000) * 100, dtype="int32")
    numpy_bincount = np.bincount(array)
    metacells_bincount = ut.bincount_vector(array)
    assert np.all(numpy_bincount == metacells_bincount)

    numpy_bincount = np.bincount(array, minlength=110)
    metacells_bincount = ut.bincount_vector(array, minlength=110)
    assert np.all(numpy_bincount == metacells_bincount)


def test_most_frequent_vector() -> None:
    array = np.array(["a", "b", "a", "c"])
    assert ut.most_frequent(array) == "a"


def test_freeze_dense() -> None:
    array = np.arange(10)

    assert not ut.frozen(array)
    array[0] = -1
    assert array[0] == -1

    ut.freeze(array)
    assert ut.frozen(array)
    try:
        array[0] = -2
        assert False, "writing to frozen array"
    except ValueError as exception:
        assert "is read-only" in str(exception)

    assert array[0] == -1

    ut.unfreeze(array)
    assert not ut.frozen(array)
    array[0] = -3
    assert array[0] == -3


def test_freeze_sparse() -> None:
    matrix = sparse.rand(100, 100, density=0.1, format="csr")

    row = 0
    while matrix.indptr[row + 1] == 0:
        row += 1
    column = matrix.indices[0]
    assert matrix[row, column] != 0

    assert not ut.frozen(matrix)
    matrix.data[0] = -1
    assert matrix[row, column] == -1

    ut.freeze(matrix)
    assert ut.frozen(matrix)
    try:
        matrix.data[0] = -2
        assert False, "writing to frozen matrix"
    except ValueError as exception:
        assert "is read-only" in str(exception)

    assert matrix.data[0] == -1

    ut.unfreeze(matrix)
    assert not ut.frozen(matrix)
    matrix.data[0] = -3
    assert matrix[row, column] == -3


def test_sliding_window_function() -> None:
    array = np.arange(5)
    actual_result = ut.sliding_window_function(array, function="mean", window_size=3)

    expected_result = np.array([1, 1, 2, 3, 3])
    assert np.allclose(actual_result, expected_result)

    order_by = np.array([0, 2, 1, 3, 4])
    actual_result = ut.sliding_window_function(array, function="mean", window_size=3, order_by=order_by)

    expected_result = np.array([1, 2, 1, 8 / 3, 8 / 3])
    assert np.allclose(actual_result, expected_result)


def test_patterns_matches() -> None:
    strings = ["foo", "bar", "baz"]
    actual = ut.patterns_matches("ba[a-z]?", strings)
    expected = np.array([False, True, True])
    assert np.allclose(actual, expected)


def test_bin_pack() -> None:
    size_of_elements = np.arange(10) * 8 + 1
    bin_of_elements = ut.bin_pack(size_of_elements, 106)
    bins_count = np.max(bin_of_elements) + 1
    size_of_bins: List[int] = []
    elements_of_bins: List[List[int]] = []
    for bin_index in range(bins_count):
        elements_of_bin = np.where(bin_of_elements == bin_index)[0]
        size_of_bin = np.sum(size_of_elements[elements_of_bin])
        assert size_of_bin <= 106
        size_of_bins.append(size_of_bin)
        elements_of_bins.append(list(elements_of_bin))

    assert size_of_bins == [98, 98, 98, 76]
    assert elements_of_bins == [[3, 9], [4, 8], [5, 7], [0, 1, 2, 6]]


def test_bin_fill() -> None:
    size_of_elements = np.arange(10) * 8 + 1
    bin_of_elements = ut.bin_fill(size_of_elements, 106)
    bins_count = np.max(bin_of_elements) + 1
    size_of_bins: List[int] = []
    elements_of_bins: List[List[int]] = []
    for bin_index in range(bins_count):
        elements_of_bin = np.where(bin_of_elements == bin_index)[0]
        size_of_bin = np.sum(size_of_elements[elements_of_bin])
        assert size_of_bin >= 106
        size_of_bins.append(size_of_bin)
        elements_of_bins.append(list(elements_of_bin))

    assert size_of_bins == [131, 123, 116]
    assert elements_of_bins == [[3, 4, 9], [2, 5, 8], [0, 1, 6, 7]]


def test_compress_indices() -> None:
    assert list(ut.compress_indices(np.array([0, 3, 2]))) == [0, 2, 1]
    assert list(ut.compress_indices(np.array([0, -1, 2]))) == [0, -1, 1]


def test_parallel_map() -> None:
    @ut.timed_call("invocation")
    def invocation(index: int) -> int:
        return index

    actual = list(ut.parallel_map(invocation, 100))
    expected = list(range(100))
    assert actual == expected

    # TODO: Why does pytest coverage error trying to read these files?
    for path in glob(".coverage.*"):
        os.remove(path)


def test_sum_groups() -> None:
    expected_sums = np.array([[5, 7, 2], [10, 8, 13]])
    expected_sizes = np.array([2, 2])
    groups = np.array([0, 1, 0, 1])

    dense_rows = np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0], [7, 8, 9]], dtype="float")

    results = ut.sum_groups(dense_rows, groups, per="row")
    assert results is not None
    assert np.allclose(ut.to_numpy_matrix(results[0]), expected_sums)
    assert np.allclose(results[1], expected_sizes)

    results = ut.sum_groups(dense_rows.transpose(), groups, per="column")
    assert results is not None
    assert np.allclose(ut.to_numpy_matrix(results[0]), expected_sums.transpose())
    assert np.allclose(results[1], expected_sizes)

    sparse_rows = sparse.csr_matrix(dense_rows)

    results = ut.sum_groups(sparse_rows, groups, per="row")
    assert results is not None
    assert np.allclose(ut.to_numpy_matrix(results[0]), expected_sums)
    assert np.allclose(results[1], expected_sizes)

    results = ut.sum_groups(sparse_rows.transpose(), groups, per="column")
    assert results is not None
    assert np.allclose(ut.to_numpy_matrix(results[0]), expected_sums.transpose())
    assert np.allclose(results[1], expected_sizes)


def test_shuffle_dense_rows_matrix() -> None:
    dense = np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0], [7, 8, 9]], dtype="float")
    ut.shuffle_matrix(dense, per="row", random_seed=123456)
    expected = np.array([[0, 2, 1], [3, 4, 0], [5, 0, 6], [7, 8, 9]], dtype="float")
    assert np.allclose(dense, expected)


def test_shuffle_sparse_rows_matrix() -> None:
    sparse_csr = sparse.csr_matrix(np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0], [7, 8, 9]], dtype="float"))
    ut.shuffle_matrix(sparse_csr, per="row", random_seed=123456)
    expected = np.array([[1, 0, 2], [3, 0, 4], [5, 0, 6], [7, 8, 9]], dtype="float")
    assert np.allclose(sparse_csr.todense(), expected)


def test_shuffle_sparse_columns_matrix() -> None:
    sparse_csc = sparse.csc_matrix(np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0], [7, 8, 9]], dtype="float"))
    ut.shuffle_matrix(sparse_csc, per="column", random_seed=123456)
    expected = np.array([[3, 1, 9], [0, 0, 4], [7, 6, 2], [5, 8, 0]], dtype="float")
    assert np.allclose(sparse_csc.todense(), expected)


def test_rank_matrix() -> None:
    dense = np.array([[0, 1, 2], [3, 0, 4], [5, 6, 0], [7, 8, 9]], dtype="float")

    result = ut.rank_per(dense, per="row", rank=1)
    expected = np.array([1, 3, 5, 8])
    assert np.allclose(result, expected)

    dense = ut.to_layout(dense, layout="column_major")
    result = ut.rank_per(dense, per="column", rank=1)
    expected = np.array([3, 1, 2])
    assert np.allclose(result, expected)


def test_random_piles() -> None:
    result = ut.random_piles(10, target_pile_size=3, random_seed=123456)
    expected = np.array([2, 2, 1, 1, 1, 0, 0, 2, 0, 0])
    assert np.allclose(result, expected)


def test_dense_per() -> None:
    matrix = np.array([[0, 1, 2], [3, 4, 5]], dtype="float")
    _test_per(matrix)


def test_sparse_per() -> None:
    matrix = np.array([[0, 1, 2], [3, 4, 5]], dtype="float")
    matrix = sparse.csr_matrix(matrix)
    _test_per(matrix)


def _test_per(rows_matrix: ut.Matrix) -> None:
    columns_matrix = ut.to_layout(rows_matrix, layout="column_major")

    assert np.allclose(ut.nnz_per(rows_matrix, per="row"), np.array([2, 3]))
    assert np.allclose(ut.nnz_per(columns_matrix, per="column"), np.array([1, 2, 2]))

    assert np.allclose(ut.sum_per(rows_matrix, per="row"), np.array([3, 12]))
    assert np.allclose(ut.sum_per(columns_matrix, per="column"), np.array([3, 5, 7]))

    assert np.allclose(ut.max_per(rows_matrix, per="row"), np.array([2, 5]))
    assert np.allclose(ut.max_per(columns_matrix, per="column"), np.array([3, 4, 5]))

    assert np.allclose(ut.min_per(rows_matrix, per="row"), np.array([0, 3]))
    assert np.allclose(ut.min_per(columns_matrix, per="column"), np.array([0, 1, 2]))

    assert np.allclose(ut.sum_squared_per(rows_matrix, per="row"), np.array([5, 50]))
    assert np.allclose(ut.sum_squared_per(columns_matrix, per="column"), np.array([9, 17, 29]))

    assert np.allclose(ut.fraction_per(rows_matrix, per="row"), np.array([3 / 15, 12 / 15]))
    assert np.allclose(ut.fraction_per(columns_matrix, per="column"), np.array([3 / 15, 5 / 15, 7 / 15]))

    assert np.allclose(ut.mean_per(rows_matrix, per="row"), np.array([3 / 3, 12 / 3]))
    assert np.allclose(ut.mean_per(columns_matrix, per="column"), np.array([3 / 2, 5 / 2, 7 / 2]))

    assert np.allclose(ut.median_per(rows_matrix, per="row"), np.array([3 / 3, 12 / 3]))
    assert np.allclose(ut.median_per(columns_matrix, per="column"), np.array([3 / 2, 5 / 2, 7 / 2]))

    assert np.allclose(
        ut.variance_per(rows_matrix, per="row"), np.array([5 / 3 - (3 / 3) ** 2, 50 / 3 - (12 / 3) ** 2])
    )

    assert np.allclose(
        ut.variance_per(columns_matrix, per="column"),
        np.array([9 / 2 - (3 / 2) ** 2, 17 / 2 - (5 / 2) ** 2, 29 / 2 - (7 / 2) ** 2]),
    )

    assert np.allclose(
        ut.normalized_variance_per(columns_matrix, per="column"),
        np.array(
            [(9 / 2 - (3 / 2) ** 2) / (3 / 2), (17 / 2 - (5 / 2) ** 2) / (5 / 2), (29 / 2 - (7 / 2) ** 2) / (7 / 2)]
        ),
    )

    dense = ut.to_numpy_matrix(ut.fraction_by(rows_matrix, by="row"))
    assert np.allclose(dense, np.array([[0 / 3, 1 / 3, 2 / 3], [3 / 12, 4 / 12, 5 / 12]]))

    dense = ut.to_numpy_matrix(ut.fraction_by(columns_matrix, by="column"))
    assert np.allclose(dense, np.array([[0 / 3, 1 / 5, 2 / 7], [3 / 3, 4 / 5, 5 / 7]]))



"""
Properly Sampled
----------------
"""

from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "compute_excluded_gene_umis",
    "find_properly_sampled_cells",
    "find_properly_sampled_genes",
]


def compute_excluded_gene_umis(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
) -> None:
    """
    Given an ``excluded_gene`` mask, compute the total ``excluded_umis`` of each cell.
    """
    umis_per_gene_per_cell = ut.get_vo_proper(adata, what, layout="column_major")
    excluded_genes_mask = ut.get_v_numpy(adata, "excluded_gene")
    umis_per_excluded_gene_per_cell = umis_per_gene_per_cell[:, excluded_genes_mask]
    umis_per_excluded_gene_per_cell = ut.to_layout(umis_per_excluded_gene_per_cell, layout="row_major")
    excluded_umis_per_cell = ut.sum_per(umis_per_excluded_gene_per_cell, per="row")
    ut.set_o_data(adata, "excluded_umis", excluded_umis_per_cell)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def find_properly_sampled_cells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    min_cell_total: Optional[int],
    max_cell_total: Optional[int],
    max_excluded_genes_fraction: Optional[float],
    inplace: bool = True,
) -> Optional[ut.PandasSeries]:
    """
    Detect cells with a "proper" amount of ``what`` (default: {what}) data.

    Due to both technical effects and natural variance between cells, the total number of UMIs
    varies from cell to cell. We often would like to work on cells that contain a sufficient number
    of UMIs for meaningful analysis; we sometimes also wish to exclude cells which have "too many"
    UMIs.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Observation (Cell) Annotations
        ``properly_sampled_cell``
            A boolean mask indicating whether each cell has a "proper" amount of UMIs.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas series (indexed by the observation names).

    **Computation Parameters**

    1. Exclude all cells whose total data is less than the ``min_cell_total`` (no default), unless
       it is ``None``.

    2. Exclude all cells whose total data is more than the ``max_cell_total`` (no default), unless
       it is ``None``.

    3. If ``max_excluded_genes_fraction`` (no default) is not ``None``, then exclude all cells whose sum of the excluded
       data (as defined by the ``excluded_gene`` mask) divided by the total data is more than the specified threshold.
    """
    total_umis_per_cell = ut.get_o_numpy(adata, what, sum=True)

    cells_mask = np.full(adata.n_obs, True, dtype="bool")

    if min_cell_total is not None:
        cells_mask = cells_mask & (total_umis_per_cell >= min_cell_total)

    if max_cell_total is not None:
        cells_mask = cells_mask & (total_umis_per_cell <= max_cell_total)

    if max_excluded_genes_fraction is not None:
        if not ut.has_data(adata, "excluded_umis"):
            compute_excluded_gene_umis(adata, what)
        excluded_umis_per_cell = ut.get_o_numpy(adata, "excluded_umis")
        excluded_umis_fraction_per_cell = excluded_umis_per_cell / total_umis_per_cell
        cells_mask = cells_mask & (excluded_umis_fraction_per_cell <= max_excluded_genes_fraction)

    if inplace:
        ut.set_o_data(adata, "properly_sampled_cell", cells_mask)
        return None

    ut.log_return("properly_sampled_cell", cells_mask)
    return ut.to_pandas_series(cells_mask, index=adata.obs_names)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def find_properly_sampled_genes(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    min_gene_total: int = pr.properly_sampled_min_gene_total,
    inplace: bool = True,
) -> Optional[ut.PandasSeries]:
    """
    Detect genes with a "proper" amount of ``what`` (default: {what}) data.

    Due to both technical effects and natural variance between genes, the expression of genes varies
    greatly between cells. This is exactly the information we are trying to analyze. We often would
    like to work on genes that have a sufficient level of expression for meaningful analysis.
    Specifically, it doesn't make sense to analyze genes that have zero expression in all the cells.

    .. todo::

        Provide additional optional criteria for "properly sampled genes"?

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Variable (Gene) Annotations
        ``properly_sampled_gene``
            A boolean mask indicating whether each gene has a "proper" number of UMIs.

    If ``inplace`` (default: {inplace}), this is written to the data and the function returns
    ``None``. Otherwise this is returned as a pandas series (indexed by the variable names).

    **Computation Parameters**

    1. Exclude all genes whose total data is less than the ``min_gene_total`` (default:
       {min_gene_total}).
    """
    total_of_genes = ut.get_v_numpy(adata, what, sum=True)

    genes_mask = total_of_genes >= min_gene_total

    if inplace:
        ut.set_v_data(adata, "properly_sampled_gene", genes_mask)
        return None

    ut.log_return("properly_sampled_gene", genes_mask)
    return ut.to_pandas_series(genes_mask, index=adata.obs_names)
